import json
import os
import tempfile
import time
from abc import abstractmethod
from pathlib import Path
from typing import Dict, Any

import tensorflow as tf
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping

from kolibri.data.corpus.generators import DataGenerator
from kolibri.dnn.embeddings import DefaultEmbedding
from kolibri.dnn.tasks.base_model import TaskBaseModel
from kolibri.dnn.utils import load_data_object
from kolibri.indexers.text_indexer import TextIndexer
from kolibri.logger import logger


class TextBaseModel(TaskBaseModel):

    def __init__(self, embedding=None, sequence_length=None, hyper_parameters=None, multi_label=False,
                 content_indexer=None, label_indexer=None):

        super().__init__(sequence_length, hyper_parameters, multi_label, label_indexer)

        if embedding is None:
            embedding = DefaultEmbedding()  # type: ignore
        if content_indexer is None:
            content_indexer = TextIndexer()

        self.content_indexer = content_indexer
        self.embedding = embedding

    def to_dict(self):
        model_json_str = self.tf_model.to_json()
        base_dict = super(TextBaseModel, self).to_dict()
        base_dict.update({
            'embedding': self.embedding.to_dict(),  # type: ignore
            'content_indexer': self.content_indexer.to_dict(),
            'label_indexer': self.label_indexer.to_dict(),
            'tf_model': json.loads(model_json_str)
        })
        return base_dict

    def update_hyper_parameters(self, parameters):
        if self.hyper_parameters is None:
            return

        for k, v in parameters.items():
            if k in self.hyper_parameters:
                self.hyper_parameters[k] = v

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def fit(self, x_train, y_train, x_validate=None, y_validate=None,
            *,
            batch_size: int = 64,
            epochs: int = 5,
            early_stopping_patience=3,
            callbacks=None,
            fit_kwargs: Dict = None):
        """
        Trains the model for a given number of epochs with given texts set list.
        """
        self.checkpoint_model_path = ""
        if x_validate:
            if callbacks is None:
                callbacks = []
            self.checkpoint_model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
            Path(self.checkpoint_model_path).mkdir(parents=True, exist_ok=True)
            self.checkpoint_model_path = os.path.join(self.checkpoint_model_path, 'best_weights.h5')
            callbacks.append([EarlyStopping(monitor='val_accuracy', patience=early_stopping_patience),
                              ModelCheckpoint(filepath=self.checkpoint_model_path, monitor='val_accuracy',
                                              save_best_only=True, verbose=1, mode='max')])

        train_gen = DataGenerator(x_train, y_train)
        if x_validate is not None:
            valid_gen = DataGenerator(x_validate, y_validate)
        else:
            valid_gen = None
        return self.fit_generator(train_sample_gen=train_gen,
                                  valid_sample_gen=valid_gen,
                                  batch_size=batch_size,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def build_model_generator(self, generators):
        raise NotImplementedError

    def fit_generator(self, train_sample_gen, valid_sample_gen=None,
                      *,
                      batch_size: int = 64,
                      epochs: int = 5,
                      callbacks=None,
                      fit_kwargs):

        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])

        model_summary = []
        self.tf_model.summary(print_fn=lambda x: model_summary.append(x))
        logger.debug('\n'.join(model_summary))

        train_set = train_sample_gen
        train_set.set_batch_size(batch_size)
        train_set.label_indexer = self.label_indexer
        train_set.content_indexer = self.content_indexer

        if fit_kwargs is None:
            fit_kwargs = {}

        if valid_sample_gen:
            valid_gen = train_sample_gen
            valid_gen.set_batch_size(batch_size)
            valid_gen.label_indexer = self.label_indexer
            valid_gen.content_indexer = self.content_indexer

            fit_kwargs['validation_data'] = valid_gen
            fit_kwargs['validation_steps'] = len(valid_gen)

        history = self.tf_model.fit(train_set,
                                    steps_per_epoch=len(train_set),
                                    epochs=epochs,
                                    callbacks=callbacks,
                                    **fit_kwargs)
        if os.path.exists(self.checkpoint_model_path):
            self.tf_model.load_weights(self.checkpoint_model_path)

        return self.tf_model

    @classmethod
    def load_model(cls, model_path):
        from kolibri.dnn.layers.crf import ConditionalRandomField
        model_config_path = os.path.join(model_path, 'model_config.json')
        with open(model_config_path, 'r') as f:
            model_config = json.loads(f.read())
        model = load_data_object(model_config)

        model.embedding = load_data_object(model_config['embedding'])
        model.content_indexer = load_data_object(model_config['content_indexer'])
        model.label_indexer = load_data_object(model_config['label_indexer'])

        tf_model_str = json.dumps(model_config['tf_model'])
        model.tf_model = tf.keras.models.model_from_json(tf_model_str,
                                                         {'ConditionalRandomField': ConditionalRandomField})

        if isinstance(model.tf_model.layers[-1], ConditionalRandomField):
            model.layer_crf = model.tf_model.layers[-1]

        model.tf_model.load_weights(os.path.join(model_path, 'model_weights.h5'))
        model.embedding.embed_model.load_weights(os.path.join(model_path, 'embed_model_weights.h5'))
        return model

    @abstractmethod
    def build_model(self,
                    x_data: Any,
                    y_data: Any) -> None:
        raise NotImplementedError

    def save(self, model_path: str) -> str:
        """
        Save model
        Args:
            model_path:
        """

        super(TextBaseModel, self).save(model_path)
        self.embedding.embed_model.save_weights(os.path.join(model_path, 'embed_model_weights.h5'))
        logger.info('embedding file saved to {}'.format(os.path.abspath(model_path)))
        return model_path


if __name__ == "__main__":
    path = '/var/folders/x3/_dg9_drj42l_cc70tsqkpqrw0000gn/T/1590915853.4571211'
    m = TextBaseModel.load_model(path)
    m.tf_model.summary()



import copy as cp
import warnings
from inspect import signature

from sklearn.linear_model import SGDClassifier

from kolibri.indexers import LabelIndexer
from kolibri.kolibri_component import Component
from kolibri.metrics.measures import *

_CLASSIFIER_TYPE = "classifier"
_REGRESSOR_TYPE = "regressor"


class MultiTargetEstimator(Component):
    """ Multi-Target Estimator for classification or regression.
    Use this meta learner to make single output predictors capable of learning
    a multi output problem, by applying them individually to each output. In
    the classification context, this is the "binary relevance" estimator.

    A Multi-Target Estimator model learns to predict multiple outputs for each
    instance. The outputs may either be discrete (i.e., classification),
    or continuous (i.e., regression). This estimator takes any base learner
    (which by default is LogisticRegression) and builds a separate model
    for each output, and will distribute each instance to each model
    for individual learning and prediction.
    """

    def __init__(self, base_estimator=SGDClassifier(max_iter=100), indexer=None):
        super().__init__()
        self.base_estimator = base_estimator
        self._check_estimator_type()
        self.ensemble = None
        self.n_targets = None
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=True)

    def _check_estimator_type(self):
        if hasattr(self, "_estimator_type"):
            if self.base_estimator._estimator_type == _CLASSIFIER_TYPE:
                self._estimator_type = _CLASSIFIER_TYPE
                return
            elif self.base_estimator._estimator_type == _REGRESSOR_TYPE:
                self._estimator_type = _REGRESSOR_TYPE
                return
        warnings.warn("Unknown base-estimator type, default is classification.", RuntimeWarning)
        self._estimator_type = _CLASSIFIER_TYPE

    def __configure(self):
        self.ensemble = [cp.deepcopy(self.base_estimator) for _ in range(self.n_targets)]

    def fit(self, X, y, classes=None, sample_weight=None):
        """ Fit the model.

        Fit n-estimators, one for each learning task.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples, n_targets)
            An array-like with the label values of all samples in X.

        classes:  numpy.ndarray, optional (default=None)
            Array with all possible/known class y_values. Usage varies depending on the base estimator.
            Not used for regression.

        sample_weight: numpy.ndarray of shape (n_samples), optional (default=None)
            Samples weight. If not provided, uniform weights are assumed. Usage varies depending on the base estimator.

        Returns
        -------
        MultiTargetEstimator
            self

        """
        self.indexer.build_label_dict([[cl] for cl in classes])
        y = np.array(self.indexer.numerize_label_sequences(y))

        N, L = y.shape
        self.n_targets = L
        self.__configure()

        for j in range(self.n_targets):
            if 'sample_weight' and 'classes' in signature(self.ensemble[j].fit).parameters:
                self.ensemble[j].fit(X, y[:, j], classes=classes, sample_weight=sample_weight)
            elif 'sample_weight' in signature(self.ensemble[j].fit).parameters:
                self.ensemble[j].fit(X, y[:, j], sample_weight=sample_weight)
            else:
                self.ensemble[j].fit(X, y[:, j])
        return self

    def partial_fit(self, X, y, classes=None, sample_weight=None):
        """ Partially (incrementally) fit the model.

        Partially fit each of the estimators on the X matrix and the
        corresponding y matrix.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples, n_targets)
            An array-like with the label values of all samples in X.

        classes: numpy.ndarray, optional (default=None)
            Array with all possible/known class y_values. This is an optional parameter, except
            for the first partial_fit call where it is compulsory. Not used for regression.

        sample_weight: numpy.ndarray of shape (n_samples), optional (default=None)
            Samples weight. If not provided, uniform weights are assumed. Usage varies depending on the base estimator.


        Returns
        -------
        MultiTargetEstimator
            self

        """
        if self.n_targets is None:
            # This is the first time that the model is fit
            self.fit(X=X, y=y, classes=classes, sample_weight=sample_weight)
            return self

        y = np.array(self.indexer.numerize_label_sequences(y))

        N, self.n_targets = y.shape

        if self.ensemble is None:
            self.__configure()

        for j in range(self.n_targets):
            if 'sample_weight' and 'classes' in signature(self.ensemble[j].partial_fit).parameters:
                self.ensemble[j].partial_fit(X, y[:, j], classes=classes, sample_weight=sample_weight)
            elif 'sample_weight' in signature(self.ensemble[j].partial_fit).parameters:
                self.ensemble[j].partial_fit(X, y[:, j], sample_weight=sample_weight)
            else:
                self.ensemble[j].partial_fit(X, y[:, j])

        return self

    def predict(self, X, return_labels=True):
        """ Predict label values for the passed texts.

        Iterates over all the estimators, predicting with each one, to obtain
        the multi output prediction.
        
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the label values for.
            
        Returns
        -------
        numpy.ndarray
            numpy.ndarray of shape (n_samples, n_targets)
            All the predictions for the samples in X.
        """

        N, D = X.shape
        predictions = np.zeros((N, self.n_targets))
        for j in range(self.n_targets):
            predictions[:, j] = self.ensemble[j].predict(X)
        if return_labels:
            return self.indexer.inverse_transform(predictions)
        return predictions

    def predict_proba(self, X):
        """ Estimates the probability of each sample in X belonging to each of
        the existing y_values for each of the classification tasks.
        
        It's a simple call to all of the classifier's predict_proba function, 
        return the probabilities for all the classification problems.

        Not applicable for regression tasks.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the class y_values for.

        Returns
        -------
        numpy.ndarray
            An array of shape (n_samples, n_classification_tasks, n_labels), in which 
            we store the probability that each sample in X belongs to each of the y_values,
            in each of the classification tasks.
        
        """
        N, D = X.shape
        proba = np.zeros((N, self.n_targets))
        for j in range(self.n_targets):
            try:
                proba[:, j] = self.ensemble[j].predict_proba(X)[:, 1]
            except NotImplementedError or AttributeError:
                raise AttributeError("Estimator {} has no predict_proba method".format(
                    type(self.base_estimator)))
        return proba

    def reset(self):
        self.ensemble = None
        self.n_targets = None

    def _more_tags(self):
        return {'multioutput': True,
                'multioutput_only': True}



import unittest

from kolibri.pipeline import Pipeline
from kolibri.tokenizer import SentenceTokenizer, WordTokenizer


class PipeLineConsumer():

    def fit(self, xt, yt, xv=None, yv=None):
        assert len(xt) == 2
        assert yt[0] == 'class1'
        assert yt[1] == 'class2'


class TestKolibriPipeline(unittest.TestCase):

    def test_fit_generator(self):
        train_x = ["This is a text with 2 sentences that should be splitted. This is the second sentence in this text",
                   "And this is a text with 2 sentences that should be splitted. And this is the second sentence in this text"]
        train_y = ["class1", "class2"]

        pipeline = Pipeline([('sentence_splitter', SentenceTokenizer()), ('word_tokenizer', WordTokenizer()),
                             ('consumer', PipeLineConsumer())])

        res = pipeline.fit(train_x, train_y)

        print(res)


if __name__ == "__main__":
    pass



from kolibri.document import Document
from kolibri.stopwords import get_stop_words
from kolibri.tokenizer import RegexpTokenizer
from kolibri.tokenizer.tokenizer import Tokenizer

pattern = r'\b(?:[A-Z][@\.A-Z-]{2,}|[\w-]{2,}|(?:\d+(?:[\.,:]\d+)*))\b'

tknzr = RegexpTokenizer(pattern)


class WordTokenizer(Tokenizer):
    name = "word_tokenizer"

    provides = ["tokens"]
    defaults = {
        "remove-stopwords": True,
        "language": 'en',
        "lower-case": False
    }

    def __init__(self, config={}):
        super().__init__(config)
        self.stopwords = None
        self.remove_stopwords = self.component_config["remove-stopwords"]
        if self.remove_stopwords:
            self.stopwords = get_stop_words(self.component_config['language'])

    def fit(self, training_data, target):
        return self

    def tokenize(self, text):
        if self.remove_stopwords:
            if not self.component_config['lower-case']:
                return [w for w in tknzr.tokenize(text) if w.lower() not in self.stopwords]
            else:
                return [w.lower() for w in tknzr.tokenize(text) if w.lower() not in self.stopwords]

        else:
            return tknzr.tokenize(text)

    def transform(self, texts, **kwargs):
        return [self.tokenize(d) for d in texts]

    def process(self, document: Document, **kwargs):
        if hasattr(document, '_sentences'):
            document.tokens = []
            for sentence in document.sentences:
                if self.remove_stopwords:
                    document.tokens.append([w for w in tknzr.tokenize(sentence) if w not in self.stopwords])
                else:
                    document.tokens.append(tknzr.tokenize(sentence))
            document.sentences = None
        else:
            if self.remove_stopwords:
                document.tokens = [w for w in tknzr.tokenize(document.text) if w not in self.stopwords]
            else:
                document.tokens = tknzr.tokenize(document.text)
            document.raw_text = None

    def get_info(self):
        return "word_tokenizer"



import collections

import numpy as np
import sklearn
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from scipy.sparse import issparse
from scipy.sparse import vstack
from sklearn.base import clone
from sklearn.base import is_regressor
from sklearn.ensemble import AdaBoostClassifier
from sklearn.metrics import accuracy_score, classification_report, confusion_matrix, f1_score, recall_score
from sklearn.preprocessing import normalize
from sklearn.utils import check_X_y
from sklearn.utils import check_array
from sklearn.utils import check_random_state
from sklearn.utils import resample

from kolibri.indexers import LabelIndexer
from kolibri.kolibri_component import Component
from kolibri.tasks.classification.models import get_model


class FinalMeta(type(AdaBoostClassifier), type(Component)):
    pass


class SMOTEBoost(AdaBoostClassifier, Component, metaclass=FinalMeta):
    name = 'smote_boost_classifier'

    """Implementation of SMOTEBoost.
    SMOTEBoost introduces texts sampling into the AdaBoost algorithm by
    oversampling the minority class using SMOTE on each boosting iteration.  
    """

    def __init__(self,
                 component_config=None,
                 new_samples=100,
                 base_estimator='linearsvm',
                 n_estimators=50,
                 learning_rate=1.,
                 algorithm='SAMME.R',
                 random_state=None,
                 sampler=SMOTE(),
                 indexer=None):

        self.new_samples = new_samples
        self.algorithm = algorithm
        self.sampler = sampler

        _estimator = get_model(base_estimator)
        if not _estimator:
            raise Exception('Could not fond "' + base_estimator + '" classification model')
        super(SMOTEBoost, self).__init__(
            base_estimator=_estimator,
            n_estimators=n_estimators,
            learning_rate=learning_rate,
            random_state=random_state)
        Component.__init__(self, component_config=component_config)
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)

    def predict(self, X):
        pred = super().predict(X)
        return self.indexer.inverse_transform(pred)

    def fit(self, X, y, sample_weight=None, minority_target=None):
        """Build a boosted classifier/regressor from the training set (X, y),
        performing SMOTE during each boosting step.
        """
        # Check that algorithm is supported.
        if self.algorithm not in ('SAMME', 'SAMME.R'):
            raise ValueError("algorithm %s is not supported" % self.algorithm)

        # Check parameters.
        if self.learning_rate <= 0:
            raise ValueError("learning_rate must be greater than zero")

        self.indexer.build_label_dict(y)
        y = self.indexer.numerize_label_sequences(y)

        X, y = check_X_y(X.toarray(), y, y_numeric=is_regressor(self))

        if sample_weight is None:
            # Initialize weights to 1 / n_samples.
            sample_weight = np.empty(X.shape[0], dtype=np.float64)
            sample_weight[:] = 1. / X.shape[0]
        else:
            sample_weight = check_array(sample_weight, ensure_2d=False)
            # Normalize existing weights.
            sample_weight = sample_weight / sample_weight.sum(dtype=np.float64)

            # Check that the sample weights sum is positive.
            if sample_weight.sum() <= 0:
                raise ValueError(
                    "Attempting to fit with a non-positive "
                    "weighted number of samples.")

        # Check parameters.
        self._validate_estimator()

        # Clear any previous fit results.
        self.estimators_ = []
        self.estimator_weights_ = np.zeros(self.n_estimators, dtype=np.float64)
        self.estimator_errors_ = np.ones(self.n_estimators, dtype=np.float64)

        random_state = check_random_state(self.random_state)

        for iboost in range(self.n_estimators):
            print('fitting estimator no ' + str(iboost))
            # SMOTE step.
            initial_count = len(y)
            X, y = self.sampler.fit_resample(X, y)

            added_samples = len(y) - initial_count
            sample_weight_syn = np.empty(added_samples, dtype=np.float64)
            sample_weight_syn[:] = 1. / X.shape[0]

            # Combine the weights.
            sample_weight = np.append(sample_weight, sample_weight_syn).reshape(-1, 1)
            sample_weight = np.squeeze(normalize(sample_weight, axis=0, norm='l1'))

            # Boosting step.
            sample_weight, estimator_weight, estimator_error = self._boost(
                iboost,
                X, y,
                sample_weight,
                random_state)

            # Early termination.
            if sample_weight is None:
                print('sample_weight: {}'.format(sample_weight))
                break

            self.estimator_weights_[iboost] = estimator_weight
            self.estimator_errors_[iboost] = estimator_error

            # Stop if error is zero.
            # if estimator_error == 0:
            #     print('error: {}'.format(estimator_error))
            #     break

            sample_weight_sum = np.sum(sample_weight)

            # Stop if the sum of sample weights has become non-positive.
            if sample_weight_sum <= 0:
                print('sample_weight_sum: {}'.format(sample_weight_sum))
                break

            if iboost < self.n_estimators - 1:
                # Normalize.
                sample_weight /= sample_weight_sum

        return self


class RUSBoost(AdaBoostClassifier, Component, metaclass=FinalMeta):
    name = 'rus_boost_classifier'

    """Implementation of SMOTEBoost.
    SMOTEBoost introduces texts sampling into the AdaBoost algorithm by
    oversampling the minority class using SMOTE on each boosting iteration.  
    """

    def __init__(self,
                 component_config=None,
                 new_samples=100,
                 base_estimator='linearsvm',
                 n_estimators=50,
                 learning_rate=1.,
                 algorithm='SAMME.R',
                 random_state=None,
                 sampler=RandomUnderSampler(),
                 indexer=None):

        self.new_samples = new_samples
        self.algorithm = algorithm
        self.sampler = sampler

        _estimator = get_model(base_estimator)
        if not _estimator:
            raise Exception('Could not fond "' + base_estimator + '" classification model')
        super(RUSBoost, self).__init__(
            base_estimator=_estimator,
            n_estimators=n_estimators,
            learning_rate=learning_rate,
            random_state=random_state)
        Component.__init__(self, component_config=component_config)
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)

    def predict(self, X):
        pred = super().predict(X)
        return self.indexer.reverse_numerize_label_sequences(pred)

    def fit(self, X, y, sample_weight=None, minority_target=None):
        """Build a boosted classifier/regressor from the training set (X, y),
        performing SMOTE during each boosting step.
        """
        # Check that algorithm is supported.
        if self.algorithm not in ('SAMME', 'SAMME.R'):
            raise ValueError("algorithm %s is not supported" % self.algorithm)

        # Check parameters.
        if self.learning_rate <= 0:
            raise ValueError("learning_rate must be greater than zero")

        self.indexer.build_label_dict(y)
        y = self.indexer.numerize_label_sequences(y)

        X, y = check_X_y(X.toarray(), y, y_numeric=is_regressor(self))

        if sample_weight is None:
            # Initialize weights to 1 / n_samples.
            sample_weight = np.empty(X.shape[0], dtype=np.float64)
            sample_weight[:] = 1. / X.shape[0]
        else:
            sample_weight = check_array(sample_weight, ensure_2d=False)
            # Normalize existing weights.
            sample_weight = sample_weight / sample_weight.sum(dtype=np.float64)

            # Check that the sample weights sum is positive.
            if sample_weight.sum() <= 0:
                raise ValueError(
                    "Attempting to fit with a non-positive "
                    "weighted number of samples.")

        # Check parameters.
        self._validate_estimator()

        # Clear any previous fit results.
        self.estimators_ = []
        self.estimator_weights_ = np.zeros(self.n_estimators, dtype=np.float64)
        self.estimator_errors_ = np.ones(self.n_estimators, dtype=np.float64)

        random_state = check_random_state(self.random_state)

        for iboost in range(self.n_estimators):
            print('fitting estimator no ' + str(iboost))
            # SMOTE step.

            X, y = self.sampler.fit_resample(X, y)

            indices = np.array(self.sampler.sample_indices_)

            sample_weight = sample_weight[indices]
            # Boosting step.
            sample_weight, estimator_weight, estimator_error = self._boost(
                iboost,
                X, y,
                sample_weight,
                random_state)

            # Early termination.
            if sample_weight is None:
                print('sample_weight: {}'.format(sample_weight))
                break

            self.estimator_weights_[iboost] = estimator_weight
            self.estimator_errors_[iboost] = estimator_error

            # Stop if error is zero.
            # if estimator_error == 0:
            #     print('error: {}'.format(estimator_error))
            #     break

            sample_weight_sum = np.sum(sample_weight)

            # Stop if the sum of sample weights has become non-positive.
            if sample_weight_sum <= 0:
                print('sample_weight_sum: {}'.format(sample_weight_sum))
                break

            if iboost < self.n_estimators - 1:
                # Normalize.
                sample_weight /= sample_weight_sum

        return self


class BalancedBagging(Component):
    name = 'balance_bagging_classifier'

    def __init__(self, component_config=None, n_estimators=10, base_estimator='linearsvm', sampler=SMOTE(),
                 indexer=None):
        super().__init__(component_config)
        self.base_estimator = base_estimator
        self.estimator = get_model(base_estimator)
        if not self.estimator:
            raise Exception('Could not find "' + base_estimator + '" classafication model')
        self.n_estimators = n_estimators
        self.model_list = []
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)
        self.sampler = sampler

        self.accuracy = "Not computed"
        self.recall = "Not computed"
        self.f1 = "Not computed"

    def fit(self, X, y, X_val=None, y_val=None):
        self.indexer.build_label_dict(y)
        y = self.indexer.numerize_label_sequences(y)

        self.model_list = []

        for ibagging in range(self.n_estimators):
            print('fitting estimateur no ' + str(ibagging))
            b = min(0.1 * ((ibagging % 10) + 1), 1)
            X_train, y_train = resample(X, y, n_samples=int(b * len(y)), stratify=y)
            X_train, y_train = self.sampler.fit_resample(X_train, y_train)
            self.estimator = get_model(self.base_estimator)
            model = self.estimator.fit(X_train, y_train)
            self.model_list.append(model)

        if X_val is not None and y_val is not None:
            pred = self.predict(X_val)

            self.accuracy = accuracy_score(y_val, pred)
            self.recall = recall_score(y_val, pred, average='micro')
            self.f1 = f1_score(y_val, pred, average='micro')

            print(classification_report(y_val, pred, target_names=list(self.indexer.idx2label.values())))
            print(confusion_matrix(y_val, pred))

        return self

    def predict_proba(self, X):
        predictions = np.array([model.predict(X) for model in self.model_list]).transpose()

        y_pred = [collections.Counter(ar).most_common()[0][0] for ar in predictions]

        return y_pred

    def predict(self, X):
        predictions = np.array([model.predict(X) for model in self.model_list]).transpose()

        y_pred = [collections.Counter(ar).most_common()[0][0] for ar in predictions]
        return self.indexer.inverse_transform(y_pred)


class BalanceCascade(Component):
    """
    The implementation of BalanceCascade.
    Hyper-parameters:
        base_estimator : scikit-learn classifier object
            optional (default=DecisionTreeClassifier)
            The base estimator from which the ensemble is built.
        n_estimators:       Number of iterations / estimators
        k_bins:             Number of hardness bins
    """
    name = 'balance_cascade_classifier'

    def __init__(self, component_config=None, base_estimator='logreg', n_estimators=10, sampler=RandomUnderSampler(),
                 random_seed=None,
                 indexer=None):
        super().__init__(component_config)
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.random_seed = random_seed
        self.model_list = []
        self.sampler = sampler
        # Will be set in the fit function
        self.feature_cols = None
        self.base_estimator = get_model(base_estimator)
        if not self.base_estimator:
            raise Exception('Could not find "' + base_estimator + '" classafication model')

        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)

    def _fit_baselearner(self, X, y):
        model = clone(self.base_estimator)
        return model.fit(X, y)

    def class_label_statistics(self, y):

        unique, counts = np.unique(y, return_counts=True)
        class_stats = dict(zip(unique, counts))

        return sorted(class_stats.items(), key=lambda x: -x[1])

    def fit(self, X, y, print_log=False, visualize=False):
        self.indexer.build_label_dict(y)
        y = np.array(self.indexer.numerize_label_sequences(y))

        class_counts = self.class_label_statistics(y)
        n_min = self.class_label_statistics(y)[-1][1]
        n_maj = self.class_label_statistics(y)[0][1]
        ir = n_min / n_maj
        keep_fp_rate = np.power(ir, 1 / (self.n_estimators - 1))

        # Algorithm start
        for ibagging in range(1, self.n_estimators):
            X_train, y_train = self.sampler.fit_resample(X, y)

            # print ('Cascade Iter: {} X_maj: {} X_rus: {} X_min: {}'.format(
            #     ibagging, len(df_maj), len(df_min), len(df_min)))
            self.model_list.append(self._fit_baselearner(X_train, y_train))
            # drop "easy" majority samples
            y_pred = self.predict_proba(X)
            indices_to_remove = []
            for cls in class_counts:
                cls_indices = np.where(y == cls[0])[0]
                ratio_to_remove = len(cls_indices) - int(keep_fp_rate * (len(cls_indices) - n_min))
                sorted_pred_class_indices = y_pred[:, 0].argsort()
                indices_to_remove.extend([i for i in sorted_pred_class_indices[:ratio_to_remove] if i in cls_indices])

            indices_to_keep = list(range(0, len(y)))
            indices_to_keep = [i for i in indices_to_keep if i not in indices_to_remove]
            X = X[indices_to_keep]
            y = y[indices_to_keep]

        return self

    def predict_proba(self, X):
        prediction = [model.predict_proba(X) for model in self.model_list]
        y_pred = np.array(prediction).mean(axis=0)
        if y_pred.ndim == 1:
            y_pred = y_pred[:, np.newaxis]
        if y_pred.shape[1] == 1:
            y_pred = np.append(1 - y_pred, y_pred, axis=1)
        return y_pred

    def predict(self, X):
        y_pred = self.predict_proba(X)
        y_pred = np.argmax(y_pred, axis=1)
        return self.indexer.reverse_numerize_label_sequences(y_pred)


class SelfPacedEnsemble(Component):
    """ Self-paced Ensemble (SPE)

    Parameters
    ----------

    base_estimator : object, optional (default=sklearn.Tree.DecisionTreeClassifier())
        The base estimator to fit on self-paced under-sampled subsets of the dataset.
        NO need to support sample weighting.
        Built-in `fit()`, `predict()`, `predict_proba()` methods are required.

    hardness_func :  function, optional
        (default=`lambda y_true, y_pred: np.absolute(y_true-y_pred)`)
        User-specified classification hardness function
            Parameters:
                y_true: 1-d array-like, shape = [n_samples]
                y_pred: 1-d array-like, shape = [n_samples]
            Returns:
                hardness: 1-d array-like, shape = [n_samples]

    n_estimators :  integer, optional (default=10)
        The number of base estimators in the ensemble.

    k_bins :        integer, optional (default=10)
        The number of hardness bins that were used to approximate hardness distribution.

    random_state :  integer / RandomState instance / None, optional (default=None)
        If integer, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by
        `numpy.random`.


    Attributes
    ----------

    base_estimator_ : estimator
        The base estimator from which the ensemble is grown.

    estimators_ : list of estimator
        The collection of fitted base estimators.

    Example:
    ```
    import numpy as np
    from sklearn import datasets
    from sklearn.tree import DecisionTreeClassifier
    from self_paced_ensemble import SelfPacedEnsemble
    from utils import make_binary_classification_target, imbalance_train_test_split

    X, y = datasets.fetch_covtype(return_X_y=True)
    y = make_binary_classification_target(y, 7, True)
    X_train, X_test, y_train, y_test = imbalance_train_test_split(
            X, y, test_size=0.2, random_state=42)

    def absolute_error(y_true, y_pred):
        # Self-defined classification hardness function
        return np.absolute(y_true - y_pred)

    spe = SelfPacedEnsemble(
        base_estimator=DecisionTreeClassifier(),
        hardness_func=absolute_error,
        n_estimators=10,
        k_bins=10,
        random_state=42,
    ).fit(
        X=X_train,
        y=y_train,
    )
    print('auc_prc_score: {}'.format(spe.score(X_test, y_test)))
    ```

    """

    name = 'paced_classifier'

    def __init__(self, component_config=None, base_estimator='logreg',
                 hardness_func=lambda y_true, y_pred: np.absolute(0.5 - y_pred), n_estimators=10, k_bins=10,
                 random_state=None, indexer=None):

        super().__init__(component_config=component_config)
        self.base_estimator_ = get_model(base_estimator)
        self.estimators_ = []
        self._hardness_func = hardness_func
        self._n_estimators = n_estimators
        self._k_bins = k_bins
        self._random_state = random_state
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)
        self.sampler = RandomUnderSampler()

    def _fit_base_estimator(self, X, y):
        """Private function used to train a single base estimator."""
        return clone(self.base_estimator_).fit(X, y)

    def _random_under_sampling(self, X, y):
        """Private function used to perform random under-sampling."""

        return self.sampler.fit_resample(X, y)

    def class_label_statistics(self, y):

        unique, counts = np.unique(y, return_counts=True)
        class_stats = dict(zip(unique, counts))

        return sorted(class_stats.items(), key=lambda x: -x[1])

    def _self_paced_under_sampling(self, X, y, maj_samples_indexes, maj_class_index, X_min, y_min, i_estimator):
        """Private function used to perform self-paced under-sampling."""

        X_maj = X[maj_samples_indexes]
        y_maj = y[maj_samples_indexes]
        # Update hardness value estimation
        y_pred_maj = self._y_pred_maj[maj_samples_indexes][:, maj_class_index]
        hardness = self._hardness_func(y_maj, y_pred_maj)

        # If hardness values are not distinguishable, perform random smapling
        if hardness.max() == hardness.min():
            X_train, y_train = self._random_under_sampling(X_maj, y_maj, X_min, y_min)
        # Else allocate majority samples into k hardness bins
        else:
            step = (hardness.max() - hardness.min()) / self._k_bins
            bins = [];
            ave_contributions = []
            for i_bins in range(self._k_bins):
                idx = (
                        (hardness >= i_bins * step + hardness.min()) &
                        (hardness < (i_bins + 1) * step + hardness.min())
                )
                # Marginal samples with highest hardness value -> kth bin
                if i_bins == (self._k_bins - 1):
                    idx = idx | (hardness == hardness.max())
                bins.append(X_maj[idx])
                ave_contributions.append(hardness[idx].mean())

            # Update self-paced factor alpha
            alpha = np.tan(np.pi * 0.5 * (i_estimator / (self._n_estimators - 1)))
            # Caculate sampling weight
            weights = 1 / (ave_contributions + alpha)
            weights[np.isnan(weights)] = 0
            # Caculate sample number from each bin
            n_sample_bins = X_min.shape[0] * weights / weights.sum()
            n_sample_bins = n_sample_bins.astype(int) + 1

            # Perform self-paced under-sampling
            sampled_bins = []
            for i_bins in range(self._k_bins):
                if min(bins[i_bins].shape[0], n_sample_bins[i_bins]) > 0:
                    np.random.seed(self._random_state)
                    idx = np.random.choice(
                        bins[i_bins].shape[0],
                        min(bins[i_bins].shape[0], n_sample_bins[i_bins]),
                        replace=False)
                    sampled_bins.append(bins[i_bins][idx])
            if issparse(sampled_bins[0]):
                X_train = vstack(sampled_bins)
            #                X_train = vstack([X_train_maj, X_min])
            else:
                X_train_maj = np.concatenate(sampled_bins, axis=0)
            #                X_train = np.concatenate([X_train_maj, X_min])

            y_train = np.full(X_train.shape[0], y_maj[0])
        #            y_train = np.concatenate([y_train_maj, y_min])

        return X_train, y_train

    def _self_paced_under_sampling_multi(self, X, y, i_estimator):
        """Private function used to perform self-paced under-sampling."""
        class_statistics = self.class_label_statistics(y)
        label_min = class_statistics[-1][0]
        label_maj = class_statistics[0][0]

        X_min = X[y == label_min]
        y_min = y[y == label_min]

        maj_indexes = np.where(y == label_maj)
        X_train, y_train = self._self_paced_under_sampling(X, y, maj_indexes, 0, X_min, y_min, i_estimator)

        for c in class_statistics[1:-1]:
            label_maj = c[0]
            maj_indexes = np.where(y == label_maj)

            X_train_, y_train_ = self._self_paced_under_sampling(X, y, maj_indexes, label_maj, X_min, y_min,
                                                                 i_estimator)
            X_train = vstack([X_train, X_train_])
            y_train = np.concatenate([y_train, y_train_])

        return vstack([X_train, X_min]), np.concatenate([y_train, y_min])

    def fit(self, X, y, label_maj=0, label_min=1):
        """Build a self-paced ensemble of estimators from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        y : array-like, shape = [n_samples]
            The label values (class y_values).

        label_maj : int, bool or float, optional (default=0)
            The majority class label, default to be negative class.

        label_min : int, bool or float, optional (default=1)
            The minority class label, default to be positive class.

        Returns
        ------
        self : object
        """

        self.indexer.build_label_dict(y)
        y = np.array(self.indexer.numerize_label_sequences(y))
        self.estimators_ = []

        # Random under-sampling in the 1st round (cold start)
        X_train, y_train = self._random_under_sampling(X, y)
        self.estimators_.append(
            self._fit_base_estimator(
                X_train, y_train))
        self._y_pred_maj = self.predict_proba(X)

        # Loop start
        for i_estimator in range(1, self._n_estimators):
            X_train, y_train = self._self_paced_under_sampling_multi(X, y, i_estimator)
            self.estimators_.append(
                self._fit_base_estimator(
                    X_train, y_train))
            # update predicted probability
            n_clf = len(self.estimators_)
            y_pred_maj_last_clf = self.estimators_[-1].predict_proba(X)
            self._y_pred_maj = (self._y_pred_maj * (n_clf - 1) + y_pred_maj_last_clf) / n_clf

        return self

    def predict_proba(self, X):
        """Predict class probabilities for X.

        The predicted class probabilities of an input sample is computed as
        the mean predicted class probabilities of the base estimators in the
        ensemble. If base estimators do not implement a ``predict_proba``
        method, then it resorts to voting and the predicted class probabilities
        of an input sample represents the proportion of estimators predicting
        each class.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        p : array of shape = [n_samples, n_classes]
            The class probabilities of the input samples.
        """
        y_pred = np.array(
            [model.predict_proba(X) for model in self.estimators_]
        ).mean(axis=0)

        return y_pred

    def predict(self, X):
        """Predict class for X.

        The predicted class of an input sample is computed as the class with
        the highest mean predicted probability. If base estimators do not
        implement a ``predict_proba`` method, then it resorts to voting.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.

        Returns
        -------
        y : array of shape = [n_samples]
            The predicted classes.
        """

        y_pred = self.predict_proba(X)
        y_pred = np.argmax(y_pred, axis=1)
        return self.indexer.inverse_transform(y_pred)

    def score(self, X, y):
        """Returns the average precision score (equivalent to the area under
        the precision-recall curve) on the given test texts and y_values.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True y_values for X.

        Returns
        -------
        score : float
            Average precision of self.predict_proba(X)[:, 1] wrt. y.
        """
        return sklearn.metrics.average_precision_score(
            y, self.predict_proba(X)[:, 1])



from kolibri.tokenizer.tokenizer import Tokenizer
from kolibri.tools.sentence_splitter import split_single


class SentenceTokenizer(Tokenizer):
    name = "sentence_tokenizer"

    provides = ["sentences"]

    def process(self, document, **kwargs):
        document.sentences = self.tokenize(document.text)
        document.raw_text = None

    def tokenize(self, text):
        sentences = split_single(text)

        return [sent.strip() for sent in sentences if len(sent.strip()) > 0]

    def transform(self, X):
        return [self.tokenize(x) for x in X]



from kolibri.application.score.score_adapter import ScorePostProcessor


class BestMatch(ScorePostProcessor):
    """
    A score adapter that returns a response based on known responses to
    the closest matches to the input document.

    :param excluded_words:
        The excluded_words parameter allows a list of words to be set that will
        prevent the score adapter from returning documents that have text
        containing any of those words. This can be useful for preventing your
        chat bot from saying swears when it is being demonstrated in front of
        an audience.
        Defaults to None
    :type excluded_words: list
    """

    def __init__(self, application, **kwargs):
        super().__init__(application, **kwargs)

        self.excluded_words = kwargs.get('excluded_words')

    def process(self, input_document, additional_response_selection_parameters=None):
        search_results = self.search_algorithm.search(input_document)

        # Search for the closest match to the input document
        for result in search_results:
            closest_match = result

            # Stop searching if a match that is close enough is found
            if result.confidence >= self.maximum_similarity_threshold:
                break

        self.application.logger.info('Using "{}" as a close match to "{}" with a confidence of {}'.format(
            closest_match.text, input_document.text, closest_match.confidence
        ))

        response = closest_match

        return response



import numpy as np


def mcut_thresholding(dataset):
    num_samps = dataset.shape[0]
    sorted_dataset = np.sort(dataset)
    diff_scores = np.diff(sorted_dataset, axis=1)
    # Get the index of the maximum differences:
    maxdiff_1 = np.argmax(diff_scores, axis=1).reshape(1, -1).T
    maxdiff_2 = maxdiff_1 + 1
    row_index = np.array(range(num_samps)).reshape(1, -1).T
    mcut = (sorted_dataset[(row_index, maxdiff_1)] + sorted_dataset[(row_index, maxdiff_2)]) / 2
    mcut = np.reshape(mcut, mcut.shape[0])
    pred_labels = np.greater(dataset, mcut[:, None]).astype(int)
    return pred_labels


def prior_prob_per_cat(training_labels, num_cats):
    num_samples = training_labels.shape[0]
    priorprobs = np.zeros((num_cats, 1), dtype='float')
    for i in range(num_cats):
        num_curr_label = training_labels[:, i].sum()
        priorprobs[i - 1] = float(num_curr_label) / num_samples
    return priorprobs


def pcut_thresholding(scores, prior_probs, x):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    pred_labels = np.zeros((set_length, num_labels))
    # Sort texts per category (low to high) and get the sorted indices:
    sorted_scores = scores.argsort(axis=0)
    for cat in range(num_labels):
        prior = prior_probs[cat]
        k = int(np.floor(prior * x * num_labels))
        # print k
        cat_scores = sorted_scores[:, cat]
        cat_scores = np.flipud(cat_scores)
        indices = cat_scores[:k]
        pred_labels[indices, cat] = 1
    return pred_labels


def rcut_thresholding(scores, threshold):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    pred_labels = np.zeros(scores.shape, dtype='uint8')
    # Get the indices of the sorted array of texts:
    sort_index = np.argsort(scores, axis=1)
    # Reverse the order of the columns to get High to Low order:
    sort_index = np.fliplr(sort_index)
    labels = sort_index[:, :threshold]
    row_index = np.array(range(set_length)).reshape(1, -1).T
    pred_labels[(row_index, labels)] = 1
    return pred_labels


def rtcut_thresholding(scores, threshold):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    pred_labels = np.zeros((set_length, num_labels), dtype='uint8')
    # Get the indices of the sorted array of texts:
    sort_index = np.argsort(scores, axis=1)
    # Reverse the order of the columns to get High to Low order:
    sort_index = np.fliplr(sort_index)
    labels = sort_index[:, :threshold]
    row_index = np.array(range(set_length)).reshape(1, -1).T
    pred_labels[(row_index, labels)] = 1
    return pred_labels


# scut_thresholding calculates the thresholds for each category
def scut_thresholding(scores, labels):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    # print set_length
    # Calculate one threshold for each label:
    thresholds = np.zeros(num_labels)
    for lbl in range(num_labels):
        scores_lbl = scores[:, lbl]
        real_lbl = labels[:, lbl].astype('int')
        opt_th = 0.0
        min_mse = np.Inf
        for th in np.linspace(0, 1, 1000):
            pred_lbl = (scores_lbl >= th).astype('int')
            mse = (np.power(pred_lbl - real_lbl, 2).sum()).astype('float') / set_length
            # print "Threshold: %f - MSE: %f" % (th, mse)
            if mse < min_mse:
                min_mse = mse
                opt_th = th
        print("Label %d - Threshold: %f - Min MSE: %f" % (lbl + 1, opt_th, min_mse))
        thresholds[lbl] = opt_th
    return thresholds


def apply_thresholds(scores, thresholds):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    pred_labels = np.zeros((set_length, num_labels), dtype='uint8')
    for lbl in range(num_labels):
        scores_lbl = scores[:, lbl]
        th = thresholds[lbl]
        pred_labels[:, lbl] = (scores_lbl >= th).astype('uint8')
    return pred_labels


# scut_thresholding calculates the thresholds for each category
def scut_thresholding(scores, labels, min_threshold=0.1, scut_type=0):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    # print set_length
    # Calculate one threshold for each label:
    thresholds = np.zeros(num_labels)
    for lbl in range(num_labels):
        scores_lbl = scores[:, lbl]
        real_lbl = labels[:, lbl].astype('int')
        opt_th = 0.0
        min_mse = np.Inf
        for th in np.linspace(0, 1, 1000):
            pred_lbl = (scores_lbl >= th).astype('int')
            mse = (np.power(pred_lbl - real_lbl, 2).sum()).astype('float') / set_length
            # print "Threshold: %f - MSE: %f" % (th, mse)
            if mse < min_mse:
                min_mse = mse
                opt_th = th
        print("Label %d - Threshold: %f - Min MSE: %f" % (lbl + 1, opt_th, min_mse))
        thresholds[lbl] = opt_th
    if scut_type == 0:
        thresholds[thresholds < min_threshold] = np.Inf
    else:
        max_score_per_cat = np.max(scores, axis=0)
        thresholds[thresholds < min_threshold] = max_score_per_cat[thresholds < min_threshold]
    return thresholds


def apply_thresholds(scores, thresholds):
    num_labels = scores.shape[1]
    set_length = scores.shape[0]
    pred_labels = np.zeros((set_length, num_labels), dtype='uint8')
    for lbl in range(num_labels):
        scores_lbl = scores[:, lbl]
        th = thresholds[lbl]
        pred_labels[:, lbl] = (scores_lbl >= th).astype('uint8')
    return pred_labels



__all__ = ['Autoencoder', 'Conv1DAutoencoder', 'VariationalAutoencoder', 'DeepAutoencoder']

from kolibri.dnn.autoencoder.autoencoder import Autoencoder
from kolibri.dnn.autoencoder.conv_autoencoder import Conv1DAutoencoder
from kolibri.dnn.autoencoder.deep_autoencoder import DeepAutoencoder
from kolibri.dnn.autoencoder.variational_autoencoder import VariationalAutoencoder



from datetime import datetime

from kolibri.utils import ordered


class DocumentBase(object):
    """
    This class has shared methods used to
    normalize different document models.
    """

    document_field_names = [
        'id',
        'text',
        'search_text',
        'label',
        'tags',
        'created_at',
    ]

    extra_document_field_names = []

    def get_document_field_names(self):
        """
        Return the list of field names for the statement.
        """
        return self.document_field_names + self.extra_document_field_names

    def get_tags(self):
        """
        Return the list of tags for this statement.
        """
        return self.tags

    def add_tags(self, *tags):
        """
        Add a list of strings to the statement as tags.
        """
        self.tags.extend(tags)

    def serialize(self):
        """
        :returns: A dictionary representation of the statement object.
        :rtype: dict
        """
        data = {}
#        print(self.get_document_field_names())
        for field_name in self.get_document_field_names():
            format_method = getattr(self, 'get_{}'.format(
                field_name
            ), None)

            if format_method:
                data[field_name] = format_method()
            else:
                data[field_name] = getattr(self, field_name)

        return data


class Document(DocumentBase):


    __slots__ = (
        'id',
        'text',
        'search_text',
        'label',
        'target_text',
        'search_target_text',
        'tags',
        'created_at',
        'confidence',
        'output_properties'
    )
    def __init__(self,  **kwargs ):
        self.id = kwargs.get('id')
        self.text = kwargs.get('text', '')
        self.created_at = kwargs.get('created_at', datetime.now())
        self.label = kwargs.get('label', '')
        self.output_properties = kwargs.get('output_properties', set())
        self.tags = kwargs.pop('tags', [])
        self.search_text = kwargs.get('search_text', '')
        self.target_text= kwargs.get('target_text', '')
        self.search_target_text= kwargs.get('search_target_text', '')
        self.extra_document_field_names.extend(['target_text', 'search_target_text'])
        self.time = kwargs.get('time', None)
        self.target = kwargs.get('target', None)
        self.tokens = None

        output_properties = kwargs.get('output_properties', None)

        if output_properties:
            self.output_properties = output_properties
        else:
            self.output_properties = set()

        self.data={}



    def set_output_property(self, prop):
        self.output_properties.add(prop)

    def entities(self):
        return self.data['entities']

    def as_dict(self, only_output_properties=False):
        if only_output_properties:
            d = {key: value
                 for key, value in self.__dict__.items()
                 if key in self.output_properties}
            return dict(d, text=self.text)
        else:
            d = self
        return dict(target=d.label, text=self.text)

    def __eq__(self, other):
        if not isinstance(other, Document):
            return False
        else:
            return ((other.text, ordered(other.data)) ==
                    (self.text, ordered(self.data)))

    def __hash__(self):
        return hash((self.text, str(ordered(self.data))))

    # @property
    # def sentences(self):
    #     if self.tokens:
    #         return self.tokens
    #     else:
    #         return self._sentences
    #
    # @sentences.setter
    # def sentences(self, new_sentences):
    #     self._sentences = new_sentences

    @classmethod
    def build(cls, text, aClass=None, entities=None):
        data = {}
        if aClass:
            data["label"] = aClass
        if entities:
            data["entities"] = entities
        return cls(text)



"""This is a somewhat delicate package. It contains all registered components
and preconfigured templates.

Hence, it imports all of the components. To avoid cycles, no component should
import this in module scope."""

from typing import Any
from typing import Optional
from typing import Text
from typing import Type

# Classes of all known components. If a new component should be added,
# its class name should be listed here.
component_classes = [
]

# Mapping from a components name to its class to allow name based lookup.
registered_components = {c.my_name: c for c in component_classes}


def class_from_module_path(module_path):
    """Given the module name and path of a class, tries to retrieve the class.

    The loaded class can be used to instantiate new objects. """
    import importlib

    # load the module, will raise ImportError if module cannot be loaded
    if "." in module_path:
        module_name, _, class_name = module_path.rpartition('.')

        m = importlib.import_module(module_name)
        # get the class, will raise AttributeError if class cannot be found
        return getattr(m, class_name)
    else:
        return globals()[module_path]


def get_component_class(component_name):
    # type: (Text) -> Optional[Type[Component]]
    """Resolve component name to a registered components class."""
    if component_name not in registered_components:
        try:
            return class_from_module_path(component_name)
        except Exception:
            raise Exception(
                "Failed to find component class for '{}'. Unknown "
                "component name. Check your configured pipeline and make "
                "sure the mentioned component is not misspelled. If you "
                "are creating your own component, make sure it is either "
                "listed as part of the `component_classes` in "
                "`kolibti.modulesy.py` or is a proper name of a class "
                "in a module.".format(component_name))
    return registered_components[component_name]


def load_component_by_name(component_name, model_dir, metadata, cached_component, **kwargs):
    """Resolves a component and calls its load method to init it based on a
    previously persisted model."""

    component_clz = get_component_class(component_name)
    return component_clz.load(model_dir, metadata, cached_component, **kwargs)


def create_component_by_name(component_name, config):
    """Resolves a component and calls it's create method to init it based on a
    previously persisted model."""

    component_clz = get_component_class(component_name)
    return component_clz.create(config)


def module_path_from_instance(inst):
    # type: (Any) -> Text
    """Return the module path of an instances class."""
    return inst.__module__ + "." + inst.__class__.__name__


def all_subclasses(cls):
    # type: (Any) -> List[Any]
    """Returns all known (imported) subclasses of a class."""

    return cls.__subclasses__() + [g for s in cls.__subclasses__()
                                   for g in all_subclasses(s)]


def register_component(Component):
    component_classes.append(Component)
    global registered_components
    registered_components = {c.name: c for c in component_classes}



# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: numeric_feature_embedding.py
# time: 2019-05-23 09:04


from typing import Union, Optional, Tuple, List

import numpy as np
from tensorflow import keras
from tensorflow.python.keras.preprocessing.sequence import pad_sequences

import kolibri.dnn
from kolibri.dnn.embeddings.base_embedding import Embedding

# from kolibri.indexers import BaseIndexer

L = keras.layers


# Todo: A better name for this class
class NumericFeaturesEmbedding(Embedding):
    """Embedding layer without pre-training, train embedding layer while training model"""

    def info(self):
        info = super(NumericFeaturesEmbedding, self).info()
        info['config'] = {
            'feature_count': self.feature_count,
            'feature_name': self.feature_name,
            'sequence_length': self.sequence_length,
            'embedding_size': self.embedding_size
        }
        return info

    def __init__(self,
                 feature_count: int,
                 feature_name: str,
                 sequence_length: Union[str, int] = 'auto',
                 embedding_size: int = None,
                 indexer=None,
                 from_saved_model: bool = False):
        """
        Init bare embedding (embedding without pre-training)

        Args:
            sequence_length: ``'auto'``, ``'variable'`` or integer. When using ``'auto'``, use the 95% of corpus length_train
                as sequence length_train. When using ``'variable'``, model input shape will set to None, which can handle
                various length_train of input, it will use the length_train of max sequence in every batch for sequence length_train.
                If using an integer, let's say ``50``, the input output sequence length_train will set to 50.
            embedding_size: Dimension of the dense embedding.
        """
        # Dummy Type
        task = kolibri.dnn.CLASSIFICATION
        if embedding_size is None:
            embedding_size = feature_count * 8
        super(NumericFeaturesEmbedding, self).__init__(task=task,
                                                       sequence_length=sequence_length,
                                                       embedding_size=embedding_size,
                                                       indexer=indexer,
                                                       from_saved_model=from_saved_model)
        self.feature_count = feature_count
        self.feature_name = feature_name
        if not from_saved_model:
            self._build_model()

    def _build_model(self, **kwargs):
        input_tensor = L.Input(shape=(self.sequence_length,),
                               name=f'input_{self.feature_name}')
        layer_embedding = L.Embedding(self.feature_count + 1,
                                      self.embedding_size,
                                      name=f'layer_embedding_{self.feature_name}')

        embedded_tensor = layer_embedding(input_tensor)
        self.embed_model = keras.Model(input_tensor, embedded_tensor)

    def analyze_corpus(self,
                       x: Union[Tuple[List[List[str]], ...], List[List[str]]],
                       y: Union[List[List[str]], List[str]]):
        pass

    def process_x_dataset(self,
                          data: List[List[str]],
                          subset: Optional[List[int]] = None) -> Tuple[np.ndarray, ...]:
        """
        batch process feature texts while training

        Args:
            data: label dataset
            subset: subset index list

        Returns:
            vectorized feature tensor
        """
        if subset is not None:
            numerized_samples = kolibri.dnn.utils.get_list_subset(data, subset)
        else:
            numerized_samples = data

        return pad_sequences(numerized_samples, self.sequence_length, padding='post', truncating='post')


if __name__ == "__main__":
    e = NumericFeaturesEmbedding(2, feature_name='is_bold', sequence_length=10)
    e.embed_model.summary()
    print(e.embed_one([1, 2]))
    print("Hello world")



import logging
import os
from typing import Any, Dict, Optional, Text

import joblib

from kolibri.features.features import Features
from kolibri.features.text.text import StreamedTfidfVectorizer
from kolibri.stopwords import get_stop_words

# from stop_words import get_stop_words
logger = logging.getLogger(__name__)


class StreamedTFIDFFeaturizer(Features):
    """Bag of words featurizer

    Creates bag-of-words representation of intent features
    using sklearn's `CountVectorizer`.
    All tokens which consist only of digits (e.g. 123 and 99
    but not ab12d) will be represented by a single feature."""

    name = "streamed_tf_idf_featurizer"

    provides = ["text_features"]

    requires = ["tokens"]

    defaults = {
        # the parameters are taken from
        # sklearn's CountVectorizer

        # regular expression for tokens
        "token_pattern": r'(?u)\b\w\w+\b',

        # remove accents during the preprocessing step
        "strip_accents": None,  # {'ascii', 'unicode', None}

        # list of stop words
        "stop_words": None,  # string {'en'}, list, or None (default)

        # min document frequency of a word to add to vocabulary
        # float - the parameter represents a proportion of documents
        # integer - absolute counts
        "min_df": 7,  # float in range [0.0, 1.0] or int

        # max document frequency of a word to add to vocabulary
        # float - the parameter represents a proportion of documents
        # integer - absolute counts
        "max_df": 0.7,  # float in range [0.0, 1.0] or int

        # set range of ngrams to be extracted
        "min_ngram": 1,  # int
        "max_ngram": 1,  # int

        # limit vocabulary size
        "max_features": 5000,  # int or None

        # if convert all characters to lowercase
        "lower-case": False,  # bool
        "use_bigram_model": False

    }

    @classmethod
    def required_packages(cls):
        return ["sklearn"]

    def _load_count_vect_params(self):
        # regular expression for tokens
        self.token_pattern = self.component_config['token_pattern']

        # remove accents during the preprocessing step
        self.strip_accents = self.component_config['strip_accents']

        # list of stop words
        self.stop_words = self.component_config['stop_words']

        # min number of word occurancies in the document to add to vocabulary
        self.min_df = self.component_config['min_df']

        # max number (fraction if float) of word occurancies
        # in the document to add to vocabulary
        self.max_df = self.component_config['max_df']

        # set ngram range
        self.min_ngram = self.component_config['min_ngram']
        self.max_ngram = self.component_config['max_ngram']

        # limit vocabulary size
        self.max_features = self.component_config['max_features']

        # if convert all characters to lowercase
        self.lowercase = not self.component_config['case_sensitive']

    def __init__(self, component_config=None):
        """Construct a new count vectorizer using the sklearn framework."""

        super(StreamedTFIDFFeaturizer, self).__init__(component_config)

        # parameters for sklearn's CountVectorizer
        self._load_count_vect_params()
        self.use_bigram_model = self.component_config["use_bigram_model"]
        # declare class instance for CountVectorizer
        self.stop_words = get_stop_words('en')

        self.vectorizer = StreamedTfidfVectorizer(min_df=self.min_df, sublinear_tf=True, max_df=self.max_df,
                                                  tokenizer=self._identity_tokenizer, lowercase=False,
                                                  stop_words=self.stop_words)

    def _identity_tokenizer(self, text):
        return text

    def fit(self, training_data, y):
        """Take parameters from config and
            construct a new tfidf vectorizer using the sklearn framework."""

        self.vectorizer.fit(training_data)

        return self

    def partial_fit(self, data, y):
        self.vectorizer.fit_partial(data, y)
        self.vectorizer.finalize_partial()
        return self

    def transform(self, documents, **kwargs):

        if self.vectorizer is None:
            logger.error("There is no trained CountVectorizer: "
                         "component is either not trained or "
                         "didn't receive enough training texts")
        else:
            #            document_text = self._get_document_text(document)

            return self.vectorizer.transform(raw_documents=documents)

    def persist(self, model_dir):
        # type: (Text) -> Dict[Text, Any]
        """Persist this model into the passed directory.
        Returns the metadata necessary to load the model again."""

        featurizer_file = os.path.join(model_dir, self.name + ".pkl")
        joblib.dump(self, featurizer_file)
        return {"featurizer_file": self.name + ".pkl"}

    @classmethod
    def load(cls,
             model_dir=None,  # type: Text
             model_metadata=None,  # type: Metadata
             cached_component=None,  # type: Optional[Component]
             **kwargs  # type: Any
             ):
        # type: (...) -> TFIDFFeaturizer

        meta = model_metadata.for_component(cls.my_name)

        if model_dir and meta.get("featurizer_file"):
            file_name = meta.get("featurizer_file")
            featurizer_file = os.path.join(model_dir, file_name)
            return joblib.load(featurizer_file)
        else:
            logger.warning("Failed to load featurizer. Maybe path {} "
                           "doesn't exist".format(os.path.abspath(model_dir)))
            return StreamedTFIDFFeaturizer(meta)



import logging

import numpy as np
from sklearn.utils import class_weight

from kolibri import settings
from kolibri.dnn.tasks.audio.classification.models import get_model
from kolibri.dnn.tasks.dnn_estimator import DnnEstimator
from kolibri.indexers.label_indexer import LabelIndexer

logger = logging.getLogger(__name__)

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"
DNN_MODEL_FILE_NAME = "classifier_dnn"


class DnnAudioEstimator(DnnEstimator):
    """classifier using the sklearn framework"""

    _estimator_type = 'estimator'

    name = 'dnn_audio_classifier'

    provides = ["classification", "target_ranking"]

    requires = []

    defaults = {

        # the models used in the classifier if several models are given they will be combined
        "models": "conv_1d",
        "multi-label": False,
        "epochs": 10,
        "loss": 'categorical_crossentropy',
        "class-weight": False,
        "project-dir": "."
    }

    def __init__(self, component_config=None):

        """Construct a new class classifier using the sklearn framework."""

        super().__init__(component_config)
        self.indexer = LabelIndexer(multi_label=self.component_config["multi-label"])

        self.clf = get_model(self.component_config['models'], hyper_parameters=self.component_config)

#        print(self.clf.tf_model.summary())
    @classmethod
    def required_packages(cls):
        return ["tensorflow"]

    def fit(self, X, y, X_val=None, y_val=None):
        fit_kwargs = {}
        if self.component_config['class-weight']:
            class_weights = class_weight.compute_class_weight('balanced', np.unique(y), y)
            fit_kwargs = {"class_weight": class_weights}
        self.clf.fit(X, y, x_validate=X_val, y_validate=y_val, epochs=self.component_config["epochs"],
                     fit_kwargs=fit_kwargs)
        self.classifier_type = type(self.clf)
    def transform(self, document):

        return self.clf.transform(document, )




    def process(self, document, **kwargs):
        """Return the most likely class and its probability for a document."""
        raw_results = None
        if not self.clf:
            # component is either not trained or didn't
            # receive enough training texts
            target = None
            target_ranking = []
        else:
            X = np.array(document.tokens)
            raw_results = self.clf.predict_top_k_class([X], top_k=settings.modeling['TARGET_RANKING_LENGTH'])

            if len(raw_results) > 0:

                target = {"name": raw_results[0]['label'], "confidence": raw_results[0]['confidence']}

                target_ranking = raw_results[0]['confidence']
            else:
                target = {"name": None, "confidence": 0.0}
                target_ranking = []

        document.label = target
        document.raw_prediction_results = raw_results
        document.set_output_property("raw_prediction_results")
        document.set_output_property("label")
        document.target_ranking = target_ranking
        document.set_output_property("target_ranking")




import matplotlib

matplotlib.use('agg')
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib import gridspec

from kolibri.classifier_evaluator.metrics import roc_margins
from kolibri.classifier_evaluator.pre_process import data_type_converter

_DEFAULT_FIGURE_SIZE = (10, 10)
_DEFAULT_COLOR_PALETTE = "jet"


def plot_trade_off_curves(roc_dict: dict, **kwargs):
    """
    plot roc margin curve;
    :param roc_dict: dict, that contains series needed for the plot; i.e.,
        set 1: given fpr and tpr:
            {model_name: {"fpr": [pd.Series, np.ndarray, list],
                          "tpr": [pd.Series, np.ndarray, list],
                          "thresholds": [pd.Series, np.ndarray, list], ...}}
        set 2: given true series and predicted_prob series:
            {model_name: {"true": [pd.Series, np.ndarray, list], "predicted_prob": [pd.Series, np.ndarray, list], ..., ...}}
    :param kwargs:
        :param figure_size: tuple, e.g., (10, 10), the size of the figure;
        :param subtitle: str, sub title given to the plot;
        :param save: [bool, str], if it is saved to an external .png file;
    :return: plt.show or saved image;
    """
    roc_element_dict = dict()
    for model_name, series_dict in roc_dict.items():
        if ('fpr' in series_dict) and ('tpr' in series_dict) and ('thresholds' in series_dict):
            roc_element_dict[model_name] = series_dict
        elif ('true' in series_dict) and ('predicted_prob' in series_dict):
            if 'pos_label' in series_dict:
                roc_margin_dict = roc_margins(true=data_type_converter(series_dict['true']),
                                              predicted_prob=data_type_converter(series_dict['predicted_prob']),
                                              pos_label=series_dict['pos_label'])
            else:
                roc_margin_dict = roc_margins(true=data_type_converter(series_dict['true']),
                                              predicted_prob=data_type_converter(series_dict['predicted_prob']))

            roc_element_dict[model_name] = roc_margin_dict
        else:
            raise KeyError(f"series dict does not contain needed series ({list(series_dict.keys())})")

    # plot set up:
    if 'figure_size' in kwargs:
        figure_size = kwargs['figure_size']
    else:
        figure_size = _DEFAULT_FIGURE_SIZE

    fig = plt.figure(figsize=figure_size)
    gs = gridspec.GridSpec(101, 1, figure=fig)  # set height ratio: 2:1
    ax0 = fig.add_subplot(gs[0:65, :])
    ax1 = fig.add_subplot(gs[68:100, :])

    # setting color
    n_model = len(roc_element_dict)
    cm_subsection = np.linspace(0, 1, n_model)
    cmap = cm.get_cmap(_DEFAULT_COLOR_PALETTE)
    colors = [cmap(x) for x in cm_subsection]

    # setting label size:
    axis_label_size = int(min(figure_size) * 1.6)
    title_label_size = int(min(figure_size) * 2.1)
    line_width = int(min(figure_size) / 6)

    ax1.set_xlabel('Thresholds', fontsize=axis_label_size)
    ax0.set_ylabel('TPR & FPR', fontsize=axis_label_size)
    ax1.set_ylabel('Margin(TPR - FPR)', fontsize=axis_label_size)

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    ax0.set_xlim([0.0, 1.0])
    ax1.set_xlim([0.0, 1.0])
    ax0.set_ylim([0.0, 1.0])
    ax1.set_ylim([0.0, 1.0])
    ax0.grid(True)

    plt.setp(ax0.set_xticks([]))
    # plt.setp(ax0.get_xticklabels(), visible=False)
    plt.subplots_adjust(hspace=.0)

    _index = -1
    for model_name, series_dict in roc_element_dict.items():
        _index += 1
        fpr = series_dict['fpr']
        tpr = series_dict['tpr']
        thresholds = series_dict['thresholds']
        margins = tpr - fpr if 'margins' not in series_dict else series_dict['margins']
        max_margin = margins.max()
        max_margin_threshold = thresholds[margins.argmax()]

        ax0.plot(thresholds, fpr, lw=line_width, linestyle='-', alpha=1,
                 label=f"Model {model_name} with TPR = {str(round(tpr[margins.argmax()], 3))} "
                       f"and FPR = {str(round(fpr[margins.argmax()], 3))}.",
                 color=colors[_index])  # fpr line;
        ax0.plot(thresholds, tpr, lw=line_width, linestyle='-', alpha=1, color=colors[_index])  # tpr line;

        ax0.axvline(x=max_margin_threshold, lw=line_width * 1.5, linestyle='-.', alpha=0.7,
                    color=colors[_index])

        # plot margins with thresholds:
        ax1.plot(thresholds, margins, lw=line_width, linestyle='-', alpha=1,
                 label=f'Model {model_name} (Max.Margin = {str(round(max_margin, 3))} at {str(round(max_margin_threshold, 3))})',
                 color=colors[_index])

        ax1.axvline(x=max_margin_threshold, lw=line_width * 1.5, linestyle='-.', alpha=0.7,
                    color=colors[_index])

    ax0.legend(loc="upper right")
    ax1.legend(loc="upper right")

    if 'subtitle' in kwargs:
        title_text = f"Trade Off Plot ({kwargs['subtitle']})"
        saved_file_name = f"trade_off_{kwargs['subtitle']}.png"
    else:
        title_text = f"Trade-Off Plot"
        saved_file_name = f"trade_off.png"

    ax0.title.set_text('Trade-Off Plot')
    ax0.title.set_fontsize(title_label_size)

    if 'save' in kwargs:
        save = kwargs['save']
        if isinstance(save, str):
            plt.savefig(save)
        else:
            plt.savefig(saved_file_name)

    plt.show()



import os

import numpy as np

from kolibri.data.corpus.base_stream import BaseStream
from kolibri.data.format.conll import read_conll
from kolibri.data.format.csv import read_csv
from kolibri.document import Document


class FileStream(BaseStream):
    """ FileStream
    A stream generated from the entries of a file.
    """
    _CLASSIFICATION = 'classification'
    _REGRESSION = 'regression'

    def __init__(self, filepath, content_col, target_idx=-1, target_cols=None, filetype='csv'):
        super().__init__()
        self.filename = ''
        self.X = None
        self.y = None
        self.content_column = content_col
        if not isinstance(target_cols, list):
            target_cols = [target_cols]
        self.target_columns = target_cols
        self.task_type = None
        self.n_classes = 0
        self.target_idx = target_idx
        self.filepath = filepath
        self.filename = ''
        self.basename = ''
        self.filetype = filetype
        if target_cols:
            self.n_targets = len(target_cols)
        # Automatically infer target_idx if not passed in multi-output problems
        if self.n_targets > 1 and self.target_idx == -1:
            self.target_idx = -self.n_targets
        self.min_class_count = 0
        self.__configure()

    def __configure(self):
        self.basename = os.path.basename(self.filepath)
        filename, extension = os.path.splitext(self.basename)
        if self.filetype == 'csv':
            self.read_function = read_csv
        elif self.filetype == 'conll':
            self.read_function = read_conll
        else:
            raise ValueError('Unsupported format: ', extension)
        self.filename = filename

    @property
    def target_idx(self):
        """
        Get the number of the column where Y begins.
        Returns
        -------
        int:
            The number of the column where Y begins.
        """
        return self._target_idx

    @target_idx.setter
    def target_idx(self, target_idx):
        """
        Sets the number of the column where Y begins.
        Parameters
        ----------
        target_idx: int
        """

        self._target_idx = target_idx

    @property
    def n_targets(self):
        """
         Get the number of targets.
        Returns
        -------
        int:
            The number of targets.
        """
        return self._n_targets

    @n_targets.setter
    def n_targets(self, n_targets):
        """
        Sets the number of targets.
        Parameters
        ----------
        n_targets: int
        """

        self._n_targets = n_targets

    @property
    def cat_features_idx(self):
        """
        Get the list of the categorical features index.
        Returns
        -------
        list:
            List of categorical features index.
        """
        return self._cat_features_idx

    @cat_features_idx.setter
    def cat_features_idx(self, cat_features_idx):
        """
        Sets the list of the categorical features index.
        Parameters
        ----------
        cat_features_idx:
            List of categorical features index.
        """

        self._cat_features_idx = cat_features_idx

    def prepare(self, min_class_count=0):
        """ prepare_for_use
        Prepares the stream for use. This functions should always be
        called after the stream initialization.
        """
        self.min_class_count = min_class_count
        self._load_data()
        self.sample_idx = 0
        self.current_sample_x = None
        self.current_sample_y = None

    def _load_data(self):
        try:
            raw_data = self.read_function(self.filepath, file_type=self.filetype)
            self.n_samples = 0
            for line in raw_data:
                if self.content_column is None:
                    self.content_column=[k for k in line if k not in self.target_columns]
                self.n_samples += 1
                if self.filetype == 'conll':
                    for token in line:
                        if self.n_targets == 1:
                            self.target_values.append(token[self.target_columns[0]])
                        else:
                            self.target_values.append([token[self.target_columns[i]] for i in range(self.n_targets)])

                else:
                    if self.n_targets == 1:
                        self.target_values.append(line[self.target_columns[0]])
                    else:
                        self.target_values.append([line[self.target_columns[i]] for i in range(self.n_targets)])

            targets, counts = np.unique(self.target_values, return_counts=True)
            self.target_values_statistics = dict(zip(targets, counts))
            self.target_values = [k for k, v in self.target_values_statistics.items() if v >= self.min_class_count]
            if np.issubdtype(np.array(self.target_values).dtype, np.integer) or np.issubdtype(
                    np.array(self.target_values).dtype, np.object):
                self.task_type = self._CLASSIFICATION
                self.n_classes = len(self.target_values)
            else:
                self.task_type = self._REGRESSION

            self.raw_data = self.read_function(self.filepath, file_type=self.filetype)

        except FileNotFoundError:
            raise FileNotFoundError("File {} does not exist.".format(self.filepath))
        pass

    def restart(self):
        """ restart
        Restarts the stream's sample feeding, while keeping all of its
        parameters.
        It basically server the purpose of reinitializing the stream to
        its initial state.
        """
        self.sample_idx = 0
        self.raw_data = self.read_function(self.filepath)
        self.current_sample_x = None
        self.current_sample_y = None

    def get_data(self, nb_samples=None):
        X = []
        Y = []
        if nb_samples:
            current_size = 0
            for x, y in self:
                X.append(x)
                Y.append(y)
                if current_size >= nb_samples:
                    break
                current_size += 1

        else:
            for x, y in self:
                X.append(x)
                Y.append(y)

        return X, Y

    def get_next_documents(self, batch_size=1):

        self.current_documents = []
        self.sample_idx += batch_size
        for i, d in enumerate(self):
            if (i % batch_size == 0 and i > 0):
                return self.current_documents
            self.current_documents.append(Document(text=d[0], target=d[1]))

        return self.current_documents

    def get_all_documents(self):
        return [Document(text=x, target=y) for x, y in self]

    def next(self, batch_size=1):
        """ next_sample
        If there is enough instances to supply at least batch_size samples, those
        are returned. If there aren't a tuple of (None, None) is returned.
        Parameters
        ----------
        batch_size: int
            The number of instances to return.
        Returns
        -------
        tuple or tuple list
            Returns the next batch_size instances.
            For general purposes the return can be treated as a numpy.ndarray.
        """
        self.current_sample_x = []
        self.current_sample_y = []
        self.sample_idx += batch_size
        for i, d in enumerate(self):
            if (i % batch_size == 0 and i > 0):
                return self.current_sample_x, self.current_sample_y
            self.current_sample_x.append(d[0])
            self.current_sample_y.append(d[1])

        return self.current_sample_x, self.current_sample_y

    def __iter__(self):

        for d in self.raw_data:
            if self.n_targets == 1:
                if d[self.target_columns[0]] not in self.target_values:
                    continue
                if isinstance(self.content_column, list):
                    yield [d[k] for k in self.content_column], d[self.target_columns[0]]
                else:
                    yield d[self.content_column], d[self.target_columns[0]]

            elif self.n_targets > 1:
                if isinstance(self.content_column, list):
                    yield  [d[k] for k in self.content_column], [d[self.target_columns[i]] for i in range(self.n_targets)]

                else:
                    yield d[self.content_column], [d[self.target_columns[i]] for i in range(self.n_targets)]

    def has_more_samples(self):
        """ Checks if stream has more samples.
        Returns
        -------
        Boolean
            True if stream has more samples.
        """
        return (self.n_samples - self.sample_idx) > 0

    def n_remaining_samples(self):
        """ Returns the estimated number of remaining samples.
        Returns
        -------
        int
            Remaining number of samples.
        """
        return self.n_samples - self.sample_idx

    def print_df(self):
        """
        Prints all the samples in the stream.
        """
        print(self.X)
        print(self.y)

    def get_data_info(self):
        if self.task_type == self._CLASSIFICATION:
            return "{} - {} label(s), {} classes".format(self.basename, self.n_targets, self.n_classes)
        elif self.task_type == self._REGRESSION:
            return "{} - {} label(s)".format(self.basename, self.n_targets)

    def get_info(self):
        return 'File Stream: filename: ' + str(self.basename) + \
               '  -  n_targets: ' + str(self.n_targets)

    def save_to_directory(self, dir):

        file_name = os.path.join(dir, 'consumers-catgories.txt')
        with open(file_name, 'w') as f:
            for l in self.target_values:
                f.write(l + '\n')




from tensorflow.python import keras

from kolibri.dnn.layers.att_wgt_avg_layer import AttentionWeightedAverage, AttWgtAvgLayer
from kolibri.dnn.layers.att_wgt_avg_layer import AttentionWeightedAverageLayer
from kolibri.dnn.layers.folding_layer import FoldingLayer
from kolibri.dnn.layers.kmax_pool_layer import KMaxPoolingLayer, KMaxPoolLayer, KMaxPooling
from kolibri.dnn.layers.non_masking_layer import NonMaskingLayer
from kapre.time_frequency import Melspectrogram
from kapre.utils import Normalization2D
L = keras.layers

if __name__ == "__main__":
    print("Hello world")



import tensorflow as tf
from tensorflow.python import keras

custom_objects = tf.keras.utils.get_custom_objects()

L = tf.keras.layers
initializers = keras.initializers
InputSpec = L.InputSpec


class FoldingLayer(L.Layer):

    def __init__(self, **kwargs):
        super(FoldingLayer, self).__init__(**kwargs)
        self.input_spec = InputSpec(ndim=3)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[1], int(input_shape[2] / 2))

    def call(self, x):
        input_shape = x.get_shape().as_list()

        # split the tensor along dimension 2 into dimension_axis_size/2
        # which will give us 2 tensors
        splits = tf.split(
            x, num_or_size_splits=int(
                input_shape[2] / 2), axis=2)

        # reduce sums of the pair of rows we have split onto
        reduce_sums = [tf.reduce_sum(split, axis=2) for split in splits]

        # stack them up along the same axis we have reduced
        row_reduced = tf.stack(reduce_sums, axis=2)
        return row_reduced



from typing import Union, Optional

import numpy as np
import pandas as pd

from kolibri.classifier_evaluator.metrics import accuracy_rate_by_prob, recall_rate_by_prob, precision_rate_by_prob, \
    confusion_matrix_by_prob, confusion_matrix_by_prob_with_dual_thresholds, roc, roc_auc
from kolibri.classifier_evaluator.pre_process import data_type_converter

_DEFAULT_POS_LABEL = True

_DEFAULT_THRESHOLD = 0.5

_DEFAULT_THRESHOLDS = [0, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 1]

_TRUE_SERIES_NAME = 'true'


class OccupiedSeriesNameError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SeriesNotFound(Exception):
    def __init__(self, series_name):
        super().__init__(f"series {series_name} not found in panel.")


class ClassifierDataPanel(object):
    """
    texts panel that manages a true series and predicted prob series;
    """

    def __init__(self,
                 pos_label: Optional[Union[bool, str, int]] = None,
                 true_series: Optional[Union[pd.Series, np.ndarray, list]] = None,
                 *predicted_prob_series: Union[pd.Series, np.ndarray, list],
                 **named_predicted_prob_series: Union[pd.Series, np.ndarray, list]):
        """
        :param pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
        :param true_series: [pd.Series, np.ndarray, list, None], a series of true classes;
        :param predicted_prob_series: [pd.Series, np.ndarray, list, None],
            a series of predicted probabilities of being the positive class (without a given name);
        :param named_predicted_prob_series: [pd.Series, np.ndarray, list, None],
            a series of predicted probabilities of being the positive class (with a given name);
        """
        self.series = dict()
        self.pos_label = None

        if true_series is not None:
            self.inject_true_series(series=true_series, pos_label=pos_label)

        for series in predicted_prob_series:
            self.inject_predicted_prob(series=series, series_name=None)

        for series_name, series in named_predicted_prob_series.items():
            self.inject_predicted_prob(series=series, series_name=series_name)

    def _check_series_name(self, proposed_name: Union[str, None]):
        """
        check and proposed name for a series into the panel is valid;
        :param proposed_name: [str, None], if a proposed name for the series;
        :return: a generated name or the valid proposed name;
        """
        prefix = 'predicted_prob'
        if not proposed_name:
            suffix_index = 1
            while f"{prefix}_{suffix_index}" in self.series:
                suffix_index += 1
            proposed_name = f"{prefix}_{suffix_index}"

        elif (proposed_name == _TRUE_SERIES_NAME) or (proposed_name in self.series):
            raise OccupiedSeriesNameError(f"Series {proposed_name} is already occupied, use other name.")

        else:
            proposed_name = proposed_name

        return proposed_name

    def rename_series(self, new_names: dict):
        """
        re-naming series;
        :param new_names: dict, {previous name: new name};
        :return:
        """
        for previous_name, new_name in new_names.items():
            if previous_name in self.series:
                new_name = self._check_series_name(proposed_name=new_name)
                self.series[new_name] = self.series.pop(previous_name)

    def inject_true_series(self,
                           series: Union[pd.Series, np.ndarray, list],
                           pos_label: Union[bool, str, int] = True):
        """
        inject a true series into the panel;
        :param series: [pd.Series, np.ndarray, list, None], a series of true classes;
        :param pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
        :return:
        """
        self.series[_TRUE_SERIES_NAME] = data_type_converter(series=series)

        if pos_label is None:
            pos_label = _DEFAULT_POS_LABEL
        else:
            pass
        self.pos_label = pos_label

    def inject_predicted_prob(self,
                              series: Union[pd.Series, np.ndarray, list],
                              series_name: Optional[str] = None):
        """
        inject a predicted prob series into the panel;
        :param series: [pd.Series, np.ndarray, list, None], a series of true classes;
        :param series_name: [str, None], the name given to this series, if not given, auto generated name is given;
        :return:
        """

        series_name = self._check_series_name(proposed_name=series_name)
        self.series[series_name] = data_type_converter(series=series)

    def __add__(self, series: Union[pd.Series, np.ndarray, list]):
        """
        add a series into the panel using an operator;
        :param series: [pd.Series, np.ndarray, list, None], a series of true classes;
        :return: the panel instance;
        """
        if not self.series:
            self.inject_true_series(series=series)
        else:
            self.inject_predicted_prob(series=series,
                                       series_name=None)

        return self

    def remove_series(self, series_name: str):
        """
        remove a series from the panel;
        :param series_name: str, name of the series to be removed;
        :return:
        """
        if series_name not in self.series:
            raise SeriesNotFound(series_name)

        del self.series[series_name]


class ClassifierEvalPanel(ClassifierDataPanel):
    """
    panel that perform classifier evaluations;
    """

    def __init__(self, *args, **kwargs):
        """
        params in ClassifierDataPanel;
        :param thresholds: Optional[Union[pd.Series, np.ndarray, list]] = None,
        """
        thresholds = kwargs.pop('thresholds', None)
        pos_label = kwargs.pop('pos_label', None)

        super().__init__(pos_label, *args, **kwargs)
        self.thresholds = None
        self.set_default_thresholds(thresholds=thresholds)

    def set_default_thresholds(self, thresholds: Optional[Union[pd.Series, np.ndarray, list]] = None):
        """
        set default thresholds for confusion matrix by prob;
        :param thresholds: [pd.Series, np.ndarray, list, None], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :return:
        """
        if thresholds is None:
            thresholds = _DEFAULT_THRESHOLDS

        self.thresholds = data_type_converter(series=thresholds)

    def accuracy_rate_by_prob(self,
                              thresholds: Optional[Union[pd.Series, np.ndarray, list]] = None,
                              focus_series: Optional[Union[list, set]] = None):
        """
        compute accuracy rate for different thresholds for different predicted prob series;
        :param thresholds: [pd.Series, np.ndarray, list, None], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :return: accuracy rates, dict, {series_name: {threshold: {'Accuracy': float}}, .. };
        """
        if thresholds is None:
            thresholds = _DEFAULT_THRESHOLDS

        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            results[series_name] = dict()
            for threshold in thresholds:
                accuracy_rate = accuracy_rate_by_prob(true=self.series[_TRUE_SERIES_NAME],
                                                      predicted_prob=predicted_prob,
                                                      threshold=threshold,
                                                      pos_label=self.pos_label)

                results[series_name][threshold] = {'Accuracy': accuracy_rate}

        return results

    def recall_by_prob(self,
                       thresholds: Optional[Union[pd.Series, np.ndarray, list]] = None,
                       focus_series: Optional[Union[list, set]] = None):
        """
        compute recall rate for different thresholds for different predicted prob series;
        :param thresholds: [pd.Series, np.ndarray, list, None], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :return: recall rates, dict, {series_name: {threshold: {'Recall': float}}, .. };
        """
        if thresholds is None:
            thresholds = _DEFAULT_THRESHOLDS

        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            results[series_name] = dict()
            for threshold in thresholds:
                recall_rate = recall_rate_by_prob(true=self.series[_TRUE_SERIES_NAME],
                                                  predicted_prob=predicted_prob,
                                                  threshold=threshold,
                                                  pos_label=self.pos_label)

                results[series_name][threshold] = {'Recall': recall_rate}

        return results

    def precision_by_prob(self,
                          thresholds: Optional[list] = None,
                          focus_series: Optional[Union[list, set]] = None):
        """
        compute precision rate for different thresholds for different predicted prob series;
        :param thresholds: [pd.Series, np.ndarray, list, None], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :return: precision rates, dict, {series_name: {threshold: {'precision': float}}, .. };
        """
        if thresholds is None:
            thresholds = _DEFAULT_THRESHOLDS

        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            results[series_name] = dict()
            for threshold in thresholds:
                recall_rate = precision_rate_by_prob(true=self.series[_TRUE_SERIES_NAME],
                                                     predicted_prob=predicted_prob,
                                                     threshold=threshold,
                                                     pos_label=self.pos_label)

                results[series_name][threshold] = {'Precision': recall_rate}

        return results

    def confusion_matrix_by_prob(self,
                                 thresholds: Optional[list] = None,
                                 output_metrics: Optional[list] = None,
                                 focus_series: Optional[Union[list, set]] = None,
                                 table: bool = True, **kwargs):
        """
        confusion matrix for binary classification according to a given set of thresholds;
        :param thresholds: [pd.Series, np.ndarray, list, None], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :param output_metrics: [list, None], metrics to be outputted if selected;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :param table: bool, if exported as a pd table table;
        :param kwargs: other keyword arguments passed into
        :return: confusion matrix results, [dict, pandas.DataFrame];
        """
        if thresholds is None:
            thresholds = self.thresholds

        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            confusion_matrix_by_threshold = confusion_matrix_by_prob(true=self.series[_TRUE_SERIES_NAME],
                                                                     predicted_prob=predicted_prob,
                                                                     thresholds=thresholds,
                                                                     pos_label=self.pos_label,
                                                                     output_metrics=output_metrics,
                                                                     table=table, **kwargs)

            results[series_name] = {'metrics_by_thresholds': confusion_matrix_by_threshold}

        return results

    def confusion_matrix_by_dual_prob(self,
                                      threshold_bounds: Union[list, tuple, set],
                                      output_metrics: Optional[list] = None,
                                      unclassified_output_metrics: Optional[list] = None,
                                      focus_series: Optional[Union[list, set]] = None,
                                      table: bool = True, **kwargs):
        """
        confusion matrix for binary classification according to a given set of dual thresholds;
        :param threshold_bounds: Union[list, tuple, set], a series of thresholds over which predicted prob will
        be classified as positive, if None is given, it will use the default threshold series;
        :param output_metrics: [list, None], metrics to be outputted if selected;
        :param unclassified_output_metrics: [list, None], unclassified metrics to be outputted if selected;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :param table: bool, if exported as a pd table table;
        :param kwargs: other keyword arguments passed into
        :return: confusion matrix results, [dict, pandas.DataFrame];
        """

        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            confusion_matrix_by_threshold = confusion_matrix_by_prob_with_dual_thresholds(
                true=self.series[_TRUE_SERIES_NAME],
                predicted_prob=predicted_prob,
                threshold_bounds=threshold_bounds,
                pos_label=self.pos_label,
                output_metrics=output_metrics,
                unclassified_output_metrics=unclassified_output_metrics,
                table=table, **kwargs)

            results[series_name] = {'metrics_by_thresholds': confusion_matrix_by_threshold}

        return results

    def roc(self,
            focus_series: Optional[Union[list, set]] = None,
            auc: bool = False):
        """
        compute roc series;
        :param focus_series: [list, set], names of the predicted prob series that metrics are run on;
        :param auc: bool, if auc is also computed;
        :return: roc series, dict; {series_name: {'roc': {'fpr': series, 'tpr': series, 'thresholds': series}}, ...}
        """
        if not focus_series:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name != _TRUE_SERIES_NAME}
        else:
            selected_series = {series_name: series for series_name, series in self.series.items()
                               if series_name in focus_series}

        results = dict()
        for series_name, predicted_prob in selected_series.items():
            fpr, tpr, thresholds = roc(true=self.series[_TRUE_SERIES_NAME],
                                       predicted_prob=predicted_prob,
                                       pos_label=self.pos_label)

            results[series_name] = {'roc': {'fpr': fpr, 'tpr': tpr, 'thresholds': thresholds}}

            if auc:
                results[series_name]['auc'] = roc_auc(fpr=fpr, tpr=tpr, thresholds=thresholds)

        return results



from typing import Text


class MissingArgumentError(ValueError):
    """Raised when a function is called and not all parameters can be
    filled from the context / config.

    Attributes:
        document -- explanation of which parameter is missing
    """

    def __init__(self, document):
        # type: (Text) -> None
        super(MissingArgumentError, self).__init__(document)
        self.document = document

    def __str__(self):
        return self.document


class UnsupportedLanguageError(Exception):
    """Raised when a component is created but the language is not supported.

    Attributes:
        component -- component name
        language -- language that component doesn't support
    """

    def __init__(self, component, language):
        # type: (Text, Text) -> None
        self.component = component
        self.language = language

        super(UnsupportedLanguageError, self).__init__(component, language)

    def __str__(self):
        return "component {} does not support language {}".format(
            self.component, self.language
        )


class InvalidConfigError(ValueError):
    """Raised if an invalid configuration is encountered."""

    def __init__(self, message):
        # type: (Text) -> None
        super(InvalidConfigError, self).__init__(message)


class InvalidProjectError(Exception):
    """Raised when a model failed to load.

    Attributes:
        document -- explanation of why the model is invalid
    """

    def __init__(self, document):
        self.document = document

    def __str__(self):
        return self.document


class UnsupportedModelError(Exception):
    """Raised when a model is too old to be loaded.

    Attributes:
        document -- explanation of why the model is invalid
    """

    def __init__(self, document):
        self.document = document

    def __str__(self):
        return self.document



import logging

logger = logging.getLogger(__name__)


class IndexedTextSearch:
    name = 'indexed_text_search'

    def __init__(self, application, **kwargs):
        from kolibri.similarity import LevenshteinDistance

        self.app = application

        document_comparison_function = kwargs.get(
            'comparison_function',
            LevenshteinDistance
        )

        self.compare_documents = document_comparison_function()

        self.search_page_size = kwargs.get(
            'search_page_size', 1000
        )

    def search(self, input_document, **kwargs):
        """
        Search for best matches to the input.
        """
        logger.info('Starting search for closest text match')

        input_search_text = input_document.search_text

        if not input_document.search_text:
            logger.warn(
                'No value for search_text was available on the provided input'
            )

            input_search_text = self.app.datastore.tagger.get_text_index_string(
                input_document.text
            )

        search_parameters = {
            'search_text_contains': input_search_text,
            'page_size': self.search_page_size
        }

        if kwargs:
            search_parameters.update(kwargs)

        document_list = self.app.datastore.filter(**search_parameters)

        best_confidence_so_far = 0

        self.app.logger.info('Processing search results')

        # Find the closest matching known document
        for document in document_list:
            confidence = self.compare_documents(input_document.text, document.text)

            if confidence > best_confidence_so_far:
                best_confidence_so_far = confidence
                document.confidence = confidence

                self.app.logger.info('Similar text found: {} {}'.format(
                    document.text, confidence
                ))

                yield document


class TextSearch:
    """
    :param document_comparison_function: A comparison class.
        Defaults to ``LevenshteinDistance``.

    :param search_page_size:
        The maximum number of records to load into memory at a time when searching.
        Defaults to 1000
    """

    name = 'text_search'

    def __init__(self, application, **kwargs):
        from kolibri.similarity import LevenshteinDistance

        self.app = application

        document_comparison_function = kwargs.get(
            'document_comparison_function',
            LevenshteinDistance
        )

        self.compare_documents = document_comparison_function()

        self.search_page_size = kwargs.get(
            'search_page_size', 1000
        )

    def search(self, input_document, **additional_parameters):
        """
        Search for close matches to the input. Confidence scores for
        subsequent results will order of increasing value.

        :param input_document: A document.
        :type input_document: kolibri.document.Document

        :param **additional_parameters: Additional parameters to be passed
            to the ``filter`` method of the datastore adapter when searching.

        :rtype: Generator yielding one closest matching document at a time.
        """
        self.app.logger.info('Beginning search for close text match')

        search_parameters = {
            'persona_not_startswith': 'bot:',
            'page_size': self.search_page_size
        }

        if additional_parameters:
            search_parameters.update(additional_parameters)

        document_list = self.app.datastore.filter(**search_parameters)

        best_confidence_so_far = 0

        self.app.logger.info('Processing search results')

        # Find the closest matching known document
        for document in document_list:
            confidence = self.compare_documents(input_document, document)

            if confidence > best_confidence_so_far:
                best_confidence_so_far = confidence
                document.confidence = confidence

                self.app.logger.info('Similar text found: {} {}'.format(
                    document.text, confidence
                ))

                yield document



import string


class PosLemmaIndexTagger(object):

    def __init__(self, language='en'):
        from kolibri.nlp.language_processor import LanguageProcessor
        self.nlp = LanguageProcessor(language)
        self.language = language

        self.punctuation_table = str.maketrans(dict.fromkeys(string.punctuation))

    def get_text_index_string(self, text):
        """
        Return a string of text containing part-of-speech, lemma pairs.
        """
        bigram_pairs = []

        if len(text) <= 2:
            text_without_punctuation = text.translate(self.punctuation_table)
            if len(text_without_punctuation) >= 1:
                text = text_without_punctuation

        document = self.nlp(text)

        if len(text) <= 2:
            bigram_pairs = [
                token.lemma_.lower() for token in document
            ]
        else:
            tokens = [
                token for token in document[0].tokens if token.is_alpha and not token.is_stop
            ]

            if len(tokens) < 2:
                tokens = [
                    token for token in document[0].tokens if token.is_alpha
                ]

            for index in range(1, len(tokens)):
                bigram_pairs.append('{}:{}'.format(
                    tokens[index - 1].pos,
                    tokens[index].lemma.lower()
                ))

        if not bigram_pairs:
            bigram_pairs = [
                token.lemma.lower() for token in document[0].tokens
            ]

        return ' '.join(bigram_pairs)



import numpy as np

from kolibri.kolibri_component import Component


class Features(Component):

    def __init__(self, config):
        super().__init__(config)
        self.vectorizer = None

    @staticmethod
    def _combine_with_existing_features(document,
                                        additional_features):
        if document.get("text_features") is not None:
            return np.hstack((document.get("text_features"),
                              additional_features))
        else:
            return additional_features



import librosa
import numpy as np
import random

from kolibri.features import Features


class AudioPreprocessor(Features):
    """
    Loads and clean audio files
"""

    name = "audio_preprocessor"

    provides = ["audio_features"]

    requires = []

    defaults = {
        "envolope": True,  # compute the envolop and apply threshold

        # remove accents during the preprocessing step
        "threshold": 20,
        "sample_rate": 16000,
        "envolop": True,
        "delta_time": 1,
        "downsample": True,
        "split": True,
        "random_start": True
    }

    @classmethod
    def required_packages(cls):
        return ["librosa"]

    def __init__(self, component_config):
        """Construct a new count vectorizer using the sklearn framework."""

        super(AudioPreprocessor, self).__init__(component_config)

    def fit_transform(self, X, y):

        return self.transform(X)

    def transform(self, X):
        Xt = np.empty((len(X), 1, int(self.component_config["sample_rate"] * self.component_config["delta_time"])),
                      dtype=np.float32)

        for i, x in enumerate(X):
            if self.component_config["downsample"]:
                wav, rate = librosa.load(x, sr=self.component_config["sample_rate"])
            else:
                wav, rate = librosa.load(x)

            intervals = librosa.effects.split(wav, top_db=self.component_config["threshold"])
            wav_output = []

            wav_len = int(self.component_config["sample_rate"] * self.component_config["delta_time"])

            for sliced in intervals:
                wav_output.extend(wav[sliced[0]:sliced[1]])

            if len(wav_output) > wav_len:
                l = len(wav_output) - wav_len
                r = 0
                if self.component_config["random_start"]:
                    r = random.randint(0, l)
                wav_output = wav_output[r:wav_len + r]
            else:
                wav_output.extend(np.zeros(shape=[wav_len - len(wav_output)], dtype=np.float32))
            Xt[i,] = np.array(wav_output).reshape(1, -1)
        return Xt

    def get_info(self):
        return 'audio cleaner'



import numpy as np
from imblearn.over_sampling import RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler
from imblearn.utils import check_target_type
from sklearn.utils import check_X_y

from kolibri.utils.misc import class_label_statistics


class AutoSampler():
    """Random samples texts to bring all class frequencies into a range.
    """

    def __init__(self, min_freq=None, max_freq=None, random_state=None):
        self.min_freq = min_freq
        self.max_freq = max_freq
        self.random_state = random_state
        self.auto_balance = False

        if min_freq and max_freq:
            self.ratio = max_freq / min_freq
        else:
            self.auto_balance = True

    def fit(self, X, y):
        """Find the classes statistics before to perform sampling.
        """
        X, y = check_X_y(X, y, accept_sparse=['csr', 'csc'], dtype=None)
        y = check_target_type(y)

        if self.auto_balance:
            class_stats = class_label_statistics(y)
            std = np.std(class_stats, axis=0)
            self.min_freq = int(class_stats[-1][1] + std[1])
            self.max_freq = int(class_stats[0][1] - std[1])
            self.ratio = self.max_freq / self.min_freq

        freq = np.unique(y, return_counts=True)
        frequencies = dict(zip(freq[0], freq[1]))
        labels = list(freq[0])

        under_dict = {}
        over_dict = {}
        self.ratio_ = self.ratio
        for lbl in labels:
            count = frequencies[lbl]
            if count < self.min_freq:
                under_dict[lbl] = count
                over_dict[lbl] = self.min_freq
            elif count > self.max_freq:
                under_dict[lbl] = self.max_freq
                over_dict[lbl] = self.max_freq
            else:
                under_dict[lbl] = count
                over_dict[lbl] = count
        self.under_sampler = RandomUnderSampler(
            sampling_strategy=under_dict, random_state=self.random_state)
        self.over_sampler = RandomOverSampler(
            sampling_strategy=over_dict, random_state=self.random_state)
        return self

    def sample(self, X, y):
        """Resample the dataset.
        """
        new_X, new_y = self.under_sampler.fit_sample(X, y)
        return self.over_sampler.fit_sample(new_X, new_y)

    def fit_resample(self, X, y):
        return self.fit(X, y).sample(X, y)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="deep-kolibri",
    version="0.2.5.1",
    author="Mohamed Ben Haddou",
    author_email="mbenhaddou@mentis.io",
    include_package_data=True,
    description="Deep Learning and more NLP toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["tensorflow>=2.0"],
    url="",
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.rst', '*.json', '*.npy', '*.db'],
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        'Development Status :: 4 - Beta',
    ],
    python_requires='>=3.7',
)



import logging
import os
from typing import List
from typing import Text

import joblib
import numpy as np
from scipy.sparse import vstack
from sklearn.metrics import accuracy_score, classification_report, confusion_matrix, f1_score, recall_score

from kolibri import settings
from kolibri.data.samplers import get_sampler
from kolibri.kolibri_component import Component
from kolibri.tasks.classification.models import get_model, get_online_model

logger = logging.getLogger(__name__)

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"


def _sklearn_numpy_warning_fix():
    """Fixes unecessary warnings emitted by sklearns use of numpy.

    Sklearn will fix the warnings in their next release in ~ August 2018.

    based on https://stackoverflow.com/questions/49545947/sklearn-deprecationwarning-truth-text-of-an-array"""
    import warnings

    warnings.filterwarnings(module='sklearn*', action='ignore',
                            category=DeprecationWarning)


class Estimator(Component):
    """classifier using the sklearn framework"""

    _estimator_type = 'estimator'

    provides = ["classification", "target_ranking"]

    requires = ["text_features"]

    defaults = {

        # the models used in the classifier if several models are given they will be combined
        "models": "logreg",
        "explain": False,
        "sampler": None,
        "priors-thresolding": False
    }

    def __init__(self, component_config=None, classifier=None, indexer=None):
        from kolibri.indexers import LabelIndexer
        """Construct a new class classifier using the sklearn framework."""
        super().__init__(component_config=component_config)

        if indexer is not None:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=False)

        _sklearn_numpy_warning_fix()
        if classifier:
            clf = get_model(classifier)
            if not clf:
                clf = get_online_model(classifier)
            if clf:
                self.clf = clf
            else:
                raise Exception('Cannot find classifier ' + classifier)

        elif 'models' in self.component_config:
            self.clf = get_model(self.component_config['models'])

        if not self.clf:
            raise Exception(
                'The classification model: "' + str(self.component_config['models']) + '" could not be created.')
        self.sampler = None

        if self.component_config['sampler']:
            self.sampler = get_sampler(self.component_config['sampler'])

        self.class_priors = None
        self.accuracy = "Not computed"
        self.recall = "Not computed"
        self.f1 = "Not computed"

    @classmethod
    def required_packages(cls):
        # type: () -> List[Text]
        return ["sklearn"]

    def evaluate(self, X_val=None, y_val=None):

        if X_val is not None and y_val is not None:
            pred = self.predict(X_val)
            #            y_val=self.indexer.inverse_transform(y_val)
            self.accuracy = accuracy_score(y_val, pred)
            self.recall = recall_score(y_val, pred, average='micro')
            self.f1 = f1_score(y_val, pred, average='micro')

            print(classification_report(y_val, pred, target_names=list(self.indexer.idx2vocab.values())))
            print(confusion_matrix(y_val, pred))

    def compute_priors(self, y):
        unique, counts = np.unique(y, return_counts=True)
        self.class_priors = dict(zip(unique, counts))

        total = sum(self.class_priors.values(), 0.0)
        self.class_priors = {k: v / total for k, v in self.class_priors.items()}

    def transform(self, document):

        return self.clf.transform(document, )

    def predict_prob(self, X):
        """Given a bow vector of an input text, predict the class label.

        Return probabilities for all y_values.

        :param X: bow of input text
        :return: vector of probabilities containing one entry for each label"""

        pred_result = self.clf.predict_proba(X)
        # sort the probabilities retrieving the indices of
        # the elements in sorted order
        sorted_indices = np.fliplr(np.argsort(pred_result, axis=1))

        return pred_result, sorted_indices, pred_result[:, sorted_indices]

    def predict(self, X):
        """Given a bow vector of an input text, predict most probable label.

        Return only the most likely label.

        :param X: bow of input text
        :return: tuple of first, the most probable label and second,
                 its probability."""

        if self.component_config["priors-thresolding"]:
            predictions = None
            try:
                predictions = self.clf.predict_proba(X)
            except Exception as e:
                pass
            if not predictions is None:
                try:
                    priors = np.array([v for v in self.class_priors.values()])
                    predictions = (predictions.T - priors[:, None]) / priors[:, None]
                    predictions = np.argmax(predictions.T, axis=1)
                except Exception as e:
                    print(e)
            else:
                predictions = self.clf.predict(X)
            return self.indexer.inverse_transform(predictions)

        return self.indexer.inverse_transform(self.clf.predict(X))

    def train(self, training_data, **kwargs):

        y = self.indexer.inverse_transform([document.label for document in training_data])
        X = vstack([document.vector for document in training_data])
        self.fit(X, y)

    def process(self, document, **kwargs):
        """Return the most likely class and its probability for a document."""

        if not self.clf:
            # component is either not trained or didn't
            # receive enough training texts
            target = None
            target_ranking = []
        else:
            X = document.vector
            raw_results, class_ids, probabilities = self.predict_prob(X)
            classes = self.indexer.inverse_transform(np.ravel(class_ids))
            # `predict` returns a matrix as it is supposed
            # to work for multiple examples as well, hence we need to flatten
            classes, probabilities = classes, probabilities.flatten()

            if len(classes) > 0 and probabilities.size > 0:
                ranking = list(zip(list(classes),
                                   list(probabilities)))[:settings.modeling['TARGET_RANKING_LENGTH']]

                target = {"name": classes[0], "confidence": probabilities[0]}

                target_ranking = [{"name": class_name, "confidence": score}
                                  for class_name, score in ranking]
            else:
                target = {"name": None, "confidence": 0.0}
                target_ranking = []
        #            self.clf.classes_ = self.indexer.la

        document.label = target
        document.raw_prediction_results = raw_results
        document.set_output_property("raw_prediction_results")
        document.set_output_property("label")
        document.target_ranking = target_ranking
        document.set_output_property("target_ranking")

    @classmethod
    def load(cls, model_dir=None, model_metadata=None, cached_component=None, **kwargs):

        meta = model_metadata.for_component(cls.name)
        file_name = meta.get("classifier_file", KOLIBRI_MODEL_FILE_NAME)
        classifier_file = os.path.join(model_dir, file_name)

        if os.path.exists(classifier_file):
            model = joblib.load(classifier_file)
            return model
        else:
            return cls(meta)

    def evaluate(self, x_data, y_data, *, digits: int = 4):
        y_pred = self.predict(x_data)

        original_report = classification_report(y_data, y_pred, output_dict=True, digits=digits)
        print(classification_report(y_data, y_pred, output_dict=False, digits=digits))
        report = {
            'detail': original_report,
            **original_report['weighted avg']
        }
        return report

    def persist(self, model_dir):
        """Persist this model into the passed directory."""

        classifier_file = os.path.join(model_dir, KOLIBRI_MODEL_FILE_NAME)
        joblib.dump(self, classifier_file)

        return {
            "classifier_file": KOLIBRI_MODEL_FILE_NAME,
            "accuracy": self.accuracy,
            "recall": self.recall,
            "f1": self.f1
        }



import tensorflow as tf

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
custom_objects = tf.keras.utils.get_custom_objects()


def custom_object_scope():
    return tf.keras.utils.custom_object_scope(custom_objects)


from kolibri.dnn.__version__ import __version__

# from kolibri.dnn import layers
# from kolibri.dnn import embeddings
# from kolibri.dnn import indexers
# from kolibri.dnn import tasks
# from kolibri.dnn import utils
# from kolibri.dnn import callbacks



from kolibri.dnn.embeddings.default_embedding import DefaultEmbedding
from kolibri.dnn.embeddings.numeric_feature_embedding import NumericFeaturesEmbedding
from kolibri.dnn.embeddings.stacked_embedding import StackedEmbedding
from kolibri.dnn.embeddings.word_embedding import WordEmbedding



import json
import os
import pathlib
from abc import abstractmethod
from typing import Dict, Any

import tensorflow as tf

import kolibri
from kolibri.indexers import LabelIndexer
from kolibri.logger import logger


class TaskBaseModel(object):

    def __init__(self, sequence_length=None, hyper_parameters=None, multi_label=False, label_indexer=None):

        self.hyper_parameters = self.get_default_hyper_parameters()

        if hyper_parameters:
            self.update_hyper_parameters(hyper_parameters)

        self.tf_model: tf.keras.Model = None

        self.sequence_length: int
        self.sequence_length = sequence_length
        self.multi_label = multi_label
        if label_indexer is None:
            self.label_indexer = LabelIndexer(multi_label=multi_label)

    def to_dict(self) -> Dict[str, Any]:
        model_json_str = self.tf_model.to_json()

        return {
            'tf_version': tf.__version__,  # type: ignore
            'kolibri_version': kolibri.__version__,
            'label_indexer': self.label_indexer.to_dict(),
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'config': {
                'hyper_parameters': self.hyper_parameters,  # type: ignore
            },
            'tf_model': json.loads(model_json_str)
        }

    def update_hyper_parameters(self, parameters):
        if self.hyper_parameters is None:
            return

        for k, v in parameters.items():
            if k in self.hyper_parameters:
                self.hyper_parameters[k] = v

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def save(self, model_path: str) -> str:
        """
        Save model
        Args:
            model_path:
        """
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)
        model_path = os.path.abspath(model_path)
        with open(os.path.join(model_path, 'model_config.json'), 'w') as f:
            f.write(json.dumps(self.to_dict(), indent=2, default=str, ensure_ascii=False))
            f.close()

        self.tf_model.save_weights(os.path.join(model_path, 'model_weights.h5'))  # type: ignore
        logger.info('model saved to {}'.format(os.path.abspath(model_path)))
        return model_path

    @classmethod
    def load_model(cls, model_path):
        raise NotImplementedError

    @abstractmethod
    def build_model(self,
                    x_data: Any,
                    y_data: Any) -> None:
        raise NotImplementedError


if __name__ == "__main__":
    path = ''
    m = TaskBaseModel.load_model(path)
    m.tf_model.summary()



import re  # gex as re

from kolibri.tokenizer.tokenizer import Tokenizer


# @ray.remote
class RegexpTokenizer(Tokenizer):
    """
    A tokenizer that splits a string using a regular expression, which
    matches either the tokens or the separators between tokens.
    By default, the following flags are
        used: `re.UNICODE | re.MULTILINE | re.DOTALL`.
    """

    def __init__(self, pattern='\w+', config=None, flags=re.UNICODE | re.MULTILINE):
        # If they gave us a regexp object, extract the patterns.

        super().__init__(config)

        pattern = getattr(pattern, 'patterns', pattern)

        self._flags = flags
        self._regexp = re.compile(pattern, self._flags)

    def __hash__(self):
        return hash(self._name)

    def tokenize(self, text):
        return self._regexp.findall(str(text))


if __name__ == '__main__':
    text = """"Please add the 'Statutory > NL-Sick Leave' => See table below.
    Company
    UPI
    Legal Name - Last Name
    Preferred Name - First Name
    Type of Leave
    Start of Leave
    Estimated Last Day of Leave
    Actual Last Day of Leave
    6079 AbbVie BV Commercial
    10373417
    Bosua
    Rosanna
    Statutory > NL-Sick Leave
    29-APR-2019
    28-APR-2020
    6079 AbbVie BV Commercial
    10355526
    Scholtes
    Monique
    Statutory > NL-Sick Leave
    26-NOV-2018
    25-NOV-2019
    Thanks!
    Met vriendelijke groet"""

    tokenizer = RegexpTokenizer(r'\b(?:[A-Z][@\.A-Z-]{2,}|[\w-]{2,}|(?:\d+(?:[\.,:]\d+)*))\b')
    tokens = tokenizer.tokenize(text)
    for t in tokens:
        print(t)



from typing import List, Dict, Any

import numpy as np
from sklearn import metrics as sklearn_metrics

import kolibri
from kolibri.data.corpus.generators import DataGenerator
from kolibri.dnn.embeddings import DefaultEmbedding
from kolibri.dnn.layers import L
from kolibri.dnn.tasks.text.base_model import TextBaseModel
from kolibri.logger import logger
from kolibri.metrics.multi_label_classification import multi_label_classification_report


class BaseTextClassificationModel(TextBaseModel):
    """
    Abstract Classification Model
    """

    __task__ = 'classification'

    def to_dict(self) -> Dict:
        info = super(BaseTextClassificationModel, self).to_dict()
        info['config']['multi_label'] = self.multi_label
        return info

    def __init__(self, embedding=None, *, sequence_length=None, hyper_parameters=None, multi_label: bool = False,
                 content_indexer=None, label_indexer=None):
        """

        Args:
            embedding: embedding object
            sequence_length: target sequence length_train
            hyper_parameters: hyper_parameters to overwrite
            multi_label: is multi-label classification
            content_indexer: text processor
            label_indexer: label processor
        """
        super().__init__(embedding, sequence_length, hyper_parameters, multi_label, content_indexer, label_indexer)

    def _activation_layer(self) -> L.Layer:
        if self.multi_label:
            return L.Activation('sigmoid')
        else:
            return L.Activation('softmax')

    def build_model(self,
                    x_train,
                    y_train):
        """
        Build Model with x_data and y_data

        This function will setup a :class:`CorpusGenerator`,
         then call py:meth:`BaseTextClassificationModel.build_model_gen` for preparing processor and model

        Args:
            x_train:
            y_train:

        Returns:

        """

        train_gen = DataGenerator(x_train, y_train)
        self.build_model_generator([train_gen])

    def build_model_generator(self, generators):
        if not self.content_indexer.vocab2idx:
            self.content_indexer.build_vocab_generator(generators)

        self.label_indexer.build_vocab_generator(generators)
        self.embedding.setup_text_processor(self.content_indexer)

        if self.sequence_length is None:
            self.sequence_length = self.embedding.get_seq_length_from_corpus(generators)

        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:
        """
        Configures the model for training.
        """
        if loss is None:
            if self.multi_label:
                loss = 'binary_crossentropy'
            else:
                loss = 'sparse_categorical_crossentropy'
        if optimizer is None:
            optimizer = 'adam'
        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)

    def predict(self, x_data, *,
                top_k=5,
                batch_size: int = 32,
                truncating: bool = False,
                multi_label_threshold: float = 0.5,
                predict_kwargs: Dict = None):
        """
        Generates output predictions for the input samples.

        Computation is done in batches.

        Args:
            x_data: The input texts, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            multi_label_threshold:
            predict_kwargs: arguments passed to ``predict()`` function of ``tf.keras.Model``

        Returns:
            array(s) of predictions.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with kolibri.dnn.custom_object_scope():
            if truncating:
                seq_length = self.sequence_length
            else:
                seq_length = None
            tensor = self.content_indexer.transform(x_data, )
            logger.debug(f'predict input shape {np.array(tensor).shape} x: \n{tensor}')
            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            logger.debug(f'predict output shape {pred.shape}')
            if self.multi_label:
                multi_label_binarizer = self.label_indexer.multi_label_binarizer  # type: ignore
                res = multi_label_binarizer.inverse_transform(pred,
                                                              threshold=multi_label_threshold)
            else:
                pred_argmax = pred.argmax(-1)
                lengths = [len(sen) for sen in x_data]
                res = self.label_indexer.inverse_transform(pred_argmax,
                                                           lengths=lengths)
                logger.debug(f'predict output argmax: {pred_argmax}')

        return res

    def predict_top_k_class(self,
                            x_data,
                            top_k=5,
                            batch_size=32,
                            debug_info=False,
                            truncating: bool = False,
                            predict_kwargs: Dict = None) -> List[Dict]:
        """
        Generates output predictions with confidence for the input samples.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with kolibri.dnn.custom_object_scope():
            if truncating:
                seq_length = self.sequence_length
            else:
                seq_length = None
            tensor = self.content_indexer.transform(x_data, )

            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            new_results = []

            for sample_prob in pred:
                sample_res = zip(self.label_indexer.idx2vocab.keys(), sample_prob)
                sample_res = sorted(sample_res, key=lambda k: k[1], reverse=True)
                data = {}
                for label, confidence in sample_res[:top_k]:
                    if 'candidates' not in data:
                        if self.embedding.indexer.multi_label:
                            data['candidates'] = []
                        else:
                            data['label'] = label
                            data['confidence'] = confidence
                            data['candidates'] = []
                            continue
                    data['candidates'].append({
                        'label': label,
                        'confidence': confidence
                    })

                new_results.append(data)

            if debug_info:
                logger.info('input: {}'.format(tensor))
                logger.info('output: {}'.format(pred))
                logger.info('output argmax: {}'.format(pred.argmax(-1)))
        return new_results

    def evaluate(self, x_data, y_data, *,
                 batch_size: int = 32,
                 digits: int = 4,
                 multi_label_threshold: float = 0.5,
                 truncating: bool = False, ):
        y_pred = self.predict(x_data,
                              batch_size=batch_size,
                              truncating=truncating,
                              multi_label_threshold=multi_label_threshold)

        if self.multi_label:
            report = multi_label_classification_report(y_data,  # type: ignore
                                                       y_pred,  # type: ignore
                                                       binarizer=self.label_indexer.multi_label_binarizer)  # type: ignore

        else:
            original_report = sklearn_metrics.classification_report(y_data,
                                                                    y_pred,
                                                                    output_dict=True,
                                                                    digits=digits)
            print(sklearn_metrics.classification_report(y_data,
                                                        y_pred,
                                                        output_dict=False,
                                                        digits=digits))
            report = {
                'detail': original_report,
                **original_report['weighted avg']
            }
        return report


if __name__ == "__main__":
    pass



"""
1.hmming_distance
2.euclidean_distance
"""

import copy

import numpy as np


def hamming_distance(x, y, weights=None):
    """
    calculate hamming distance
    :param x: a sample
    :param y: another sample
    :param weights: the weights for each feature
    :return: hamming distance
    """
    assert len(x) == len(y)
    if weights is None:
        weights = np.ones(len(x))
    distance = np.sum(((1 - np.sign(x * y)) / 2) * weights)
    return distance


def euclidean_distance(x, y, weights=None):
    """
    calulate euclidean distance
    :param x: a sample
    :param y: another sample
    :param weights: the weights for each feature
    :return: euclidean distance
    """
    assert len(x) == len(y)
    if weights is None:
        weights = np.ones(len(x))
    temp_x = copy.deepcopy(x)
    temp_y = copy.deepcopy(y)
    distance = np.sqrt(np.sum(np.power(temp_x - temp_y, 2) * weights))
    return distance



import re  # To use regular expressions
import subprocess  # To run shell commands
import time

from lxml import etree  # To read an write XML files


def timeit(method):
    """Time functions."""

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        print('%r %2.2f sec' % \
              (method.__name__, te - ts))
        return result

    return timed


class freelingWrapper(object):
    """Wrap FreeLing."""

    @timeit
    def __init__(self, port):
        """Instantiate a wrapper."""

        self.port = port

    def read_xml(self, infile):
        """Parse a XML file."""
        parser = etree.XMLParser(remove_blank_text=True)
        with open(infile, encoding='utf-8', mode='r') as input:
            return etree.parse(input, parser)

    def process_with_freeling(self, itext):
        """Process a string with FreeLing.

        Keyword arguments:
        itext -- a string containing input text to be processed
        """
        itext = re.sub(r'\"', r'\\"', itext)  # scape double quotation marks
        command = 'echo "{}" | analyzer_client {}'.format(  # create command string
            itext,
            self.port)
        process = subprocess.Popen(  # declare process
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True)
        otext, error = process.communicate()  # run process
        otext = otext.decode("utf-8").strip()
        return otext

    # for CWB encoding is not a problem if the XML is pretty,
    # there is an option to cope with it at encoding
    def unprettify(self, tree):
        """Remove any indentation introduced by pretty print."""
        tree = etree.tostring(  # convert XML tree to string
            tree,
            encoding="utf-8",
            method="xml",
            xml_declaration=True).decode()
        tree = re.sub(  # remove trailing spaces before tag
            r"(\n) +(<)",
            r"\1\2",
            tree)
        tree = re.sub(  # put each XML element in a different line
            r"> *<",
            r">\n<",
            tree)
        tree = re.sub(  # put opening tag and FL output in different lines
            r"(<{}.+?>)".format(self.element),
            r"\1\n",
            tree)
        tree = re.sub(  # put FL output and closing tag in different liens
            r"(</{}>)".format(self.element),
            r"\n\1",
            tree)
        tree = re.sub(
            r"(>)([^.])",
            r"\1\n\2",
            tree)
        tree = re.sub(  # remove unnecessary empty lines
            r"\n\n+",
            r"\n",
            tree)
        return tree

    def flg_to_vrt(self, tree):
        """Transform FreeLing output to VRT."""
        if self.sentence == True:
            sentences = tree.findall('.//{}'.format(self.element))
        else:
            sentences = tree.findall('.//s')
        for sentence in sentences:
            sentence.text = re.sub(r' ', r'\t', sentence.text)  # to get VRT directly
            sentence.text = re.sub(r'\t\d(\.\d+)?$', r'', sentence.text, flags=re.MULTILINE)  # to remove probability
        pass

    def get_leafs(self, sentence):
        """Get leaf nodes from constituency tree.

        Keyword arguments:
        sentence -- XML element containing a parsed sentence
        """
        counter = 0
        nodes = sentence.find('./constituents/node')
        s = etree.Element('s')
        s.append(nodes)
        leafs = s.findall('.//node[@leaf]')
        for leaf in leafs:
            ancestors = leaf.iterancestors(tag='node')
            depth = sum([1 for x in ancestors])
            if re.match(r'.+_.+', leaf.attrib['word']):
                words = leaf.attrib['word'].split('_')
                leaf.attrib['word'] = words[0]
                leaf.attrib['token'] = 't_' + str(counter)
                leaf.attrib['depth'] = str(depth)
                leaf.text = '\n{}\n'.format(words[0])
                counter += 1
                parent = leaf.getparent()
                leaf_counter = 2
                for word in words[1:]:
                    child = etree.SubElement(parent, 'node')
                    child.text = '\n{}\n'.format(word)
                    child.attrib['leaf'] = str(leaf_counter)
                    leaf_counter += 1
                    child.attrib['token'] = 't_' + str(counter)
                    counter += 1
                    child.attrib['word'] = word
                    child.attrib['depth'] = str(depth)
                    if 'head' in leaf.attrib:
                        child.attrib['head'] = leaf.attrib['head']
            else:
                leaf.text = '\n{}\n'.format(leaf.attrib['word'])
                leaf.attrib['depth'] = str(depth)
                leaf.attrib['token'] = 't_' + str(counter)
                counter += 1
        return s

    def get_constituency(self, element):
        """Get constituency parsing of the text contained in a XML element.

        Keyword arguments:
        element -- XML element containing text to be parsed.
        """
        analysis = self.process_with_freeling(element.text.strip('\n'))
        analysis = etree.fromstring('<analysis>' + analysis.strip('\n') + '</analysis>')
        if self.sentence is True:
            sentence = analysis.find('.//sentence')
            element.text = None
            parsed_sentence = self.get_leafs(sentence)
            element.getparent().replace(element, parsed_sentence)
        else:
            sentences = analysis.findall('.//sentence')
            new_sentences = []
            for sentence in sentences:
                new_sentences.append(self.get_leafs(sentence))
            element.text = None
            for new_sentence in new_sentences:
                element.append(new_sentence)

    def process_sentences(self, sentences):
        processed = []
        for s in sentences:
            processed.append(self.process_with_freeling(
                s.strip('\n')))

        return processed

# tagger=freelingWrapper(50101)



from kolibri.meta.chain_estimator import *
from kolibri.meta.ecoc_estimator import EcocEstimator
from kolibri.meta.ensemble_samplers import *
from kolibri.meta.multi_output_estimator import MultiTargetEstimator



import logging

from kolibri.sentence import Sentence
from kolibri.stopwords import get_stop_words
from kolibri.token import Token
from kolibri.tokenizer.sentence_tokenizer import split_single
from kolibri.tools.freeling.pyfreeling import freelingWrapper


class LanguageProcessor:
    def __init__(self, language='en', logging_level=logging.WARNING):
        self.logging_level = logging_level
        self.freelang = freelingWrapper(50101)

        self.language = language
        logging.basicConfig(level=self.logging_level)
        self.stopwords = get_stop_words(self.language)

    def __enter__(self):
        return self

    def process_annotated(self, annotated):
        sentence_elements = annotated.split('\n')
        sentence = Sentence()
        for words in sentence_elements:
            element = words.split()
            token = Token()
            token.id = element[0]
            token.text = str(element[1])
            token.lemma = element[2]
            token.coarse_pos = element[3]
            token.pos = element[4]
            token.is_stop = token.text in self.stopwords
            token.is_alpha = token.is_alpha = token.text.isalpha()
            token.isnumeric = token.text.isnumeric()
            token.islower = token.text.islower()
            token.istitle = token.text.istitle()
            token.isupper = token.text.isupper()
            token.data_dic = {d.split('=')[0]: d.split('=')[1] for d in element[5].split('|')}
            sentence.tokens.append(token)

        return sentence

    def __call__(self, text, split_sentences=False):
        sentences = []
        if split_sentences:
            sents = split_single(text)
            annotated = self.freelang.process_sentences(sents)
        else:
            annotated = self.freelang.process_sentences([text])

        for ann in annotated:
            sentences.append(self.process_annotated(ann))
        return sentences



# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: base_embedding.py
# time: 2:43 下午

import json
from typing import Dict, List, Any, Optional

import numpy as np
import tensorflow as tf
import tqdm

import kolibri.dnn
from kolibri.indexers.base_indexer import BaseIndexer
from kolibri.logger import logger

L = tf.keras.layers


class Embedding:
    def to_dict(self) -> Dict[str, Any]:
        config: Dict[str, Any] = {
            'embedding_size': self.embedding_size,
            'max_position': self.max_position,
            **self.kwargs
        }
        return {
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'config': config,
            'embed_model': json.loads(self.embed_model.to_json())
        }

    def __init__(self,
                 embedding_size: int = 100,
                 max_position: int = None,
                 **kwargs: Any):

        self.embed_model: tf.keras.Model = None

        self.kwargs = kwargs

        self.embedding_size: int = embedding_size  # type: ignore
        self.max_position: int = max_position  # type: ignore
        self.vocab2idx = self.load_embed_vocab()
        self.indexer: Optional[BaseIndexer] = None

    def _override_load_model(self, config: Dict) -> None:
        embed_model_json_str = json.dumps(config['embed_model'])
        self.embed_model = tf.keras.models.model_from_json(embed_model_json_str,
                                                           custom_objects=kolibri.dnn.custom_objects)

    def setup_text_processor(self, indexer):
        self.indexer = indexer
        self.build_embedding_model(vocab_size=indexer.vocab_size)
        if self.vocab2idx:
            self.indexer.vocab2idx = self.vocab2idx
            self.indexer.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])

    def get_seq_length_from_corpus(self,
                                   generators,
                                   *,
                                   use_label: bool = False,
                                   cover_rate: float = 0.95) -> int:
        """
        Calculate proper sequence length_train according to the corpus

        Args:
            generators:
            use_label:
            cover_rate:

        Returns:

        """
        seq_lens = []
        for gen in generators:
            for sentence, label in tqdm.tqdm(gen, desc="Calculating sequence length_train"):
                if use_label:
                    seq_lens.append(len(label))
                else:
                    seq_lens.append(len(sentence))
        if cover_rate == 1.0:
            target_index = -1
        else:
            target_index = int(cover_rate * len(seq_lens))
        sequence_length = sorted(seq_lens)[target_index]
        logger.debug(f'Calculated sequence length_train = {sequence_length}')
        return sequence_length

    def load_embed_vocab(self) -> Optional[Dict[str, int]]:
        """
        Load vocab dict from embedding layer

        Returns:
            vocab dict or None
        """
        raise NotImplementedError

    def build_embedding_model(self,
                              *,
                              vocab_size: int = None,
                              force: bool = False,
                              **kwargs: Dict) -> None:
        raise NotImplementedError

    def embed(self,
              sentences: List[List[str]],
              *,
              debug: bool = False) -> np.ndarray:
        """
        batch embed sentences

        Args:
            sentences: Sentence list to embed
            debug: show debug info
        Returns:
            vectorized sentence list
        """
        if self.indexer is None:
            raise ValueError('Need to setup the `embedding.setup_text_processor` before calling the embed function.')

        tensor_x = self.indexer.transform(sentences,
                                          seq_length=self.max_position)
        if debug:
            logger.debug(f'sentence tensor: {tensor_x}')
        embed_results = self.embed_model.predict(tensor_x)
        return embed_results

if __name__ == "__main__":
    pass
